package be.iminds.ilabt.jfed.experimenter_gui.controller.view.dialogs;

import be.iminds.ilabt.jfed.experimenter_gui.controller.controller.ControllerController;
import be.iminds.ilabt.jfed.experimenter_gui.controller.view.TimeTextField;
import be.iminds.ilabt.jfed.rspec.model.controller.ExperimentBarrierSegment;
import be.iminds.ilabt.jfed.rspec.model.controller.ExperimentCommand;
import be.iminds.ilabt.jfed.rspec.model.controller.Timestamp;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecNode;
import be.iminds.ilabt.jfed.ui.javafx.FXMLUtil;
import be.iminds.ilabt.jfed.ui.javafx.GlyphUtils;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import javafx.event.ActionEvent;
import javafx.fxml.FXML;
import javafx.scene.control.CheckBoxTreeItem;
import javafx.scene.control.TextField;
import javafx.scene.control.TreeView;
import javafx.scene.control.cell.CheckBoxTreeCell;
import javafx.scene.layout.GridPane;
import javafx.scene.paint.Color;
import javafx.stage.Window;
import org.controlsfx.dialog.Dialog;
import org.controlsfx.dialog.DialogAction;
import org.controlsfx.glyphfont.FontAwesome;
import org.controlsfx.validation.ValidationMessage;
import org.controlsfx.validation.ValidationSupport;
import org.controlsfx.validation.Validator;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

public class AddMultipleCommandDialog extends GridPane {
    private static final Logger LOG = LoggerFactory.getLogger(AddMultipleCommandDialog.class);

    private final ControllerController controller;
    @FXML
    private TextField commandTextField;
    @FXML
    private TimeTextField startTextField;
    @FXML
    private TextField tagTextField;
    @FXML
    private TreeView<String> treeView;

    private Map<CheckBoxTreeItem<String>, FXRspecNode> nodesMap = new HashMap<>();
    private Map<CheckBoxTreeItem<String>, ExperimentBarrierSegment> barriersMap = new HashMap<>();

    private final org.controlsfx.dialog.Dialog dialog;
    private final DialogAction addCommandAction;

    private final ValidationSupport validationSupport;


    private AddMultipleCommandDialog(Window parent, ControllerController controller, List<FXRspecNode> nodes) {
        this.controller = controller;

        //initialize the Scene
        FXMLUtil.injectFXML(this);

        CheckBoxTreeItem<String> rootItem = new CheckBoxTreeItem<>("Your experiment");
        rootItem.setExpanded(true);
        treeView.setRoot(rootItem);
        treeView.setEditable(false);

        treeView.setCellFactory(CheckBoxTreeCell.<String>forTreeView());

        for (ExperimentBarrierSegment b : controller.getModel().getCommandModelRspec().getBarrierSegments()) {
            CheckBoxTreeItem<String> barrierItem = new CheckBoxTreeItem<>(b.getTag());
            rootItem.getChildren().add(barrierItem);
            barriersMap.put(barrierItem, b);
            for (FXRspecNode n : nodes) {
                CheckBoxTreeItem<String> cb = new CheckBoxTreeItem<>(n.getClientId());
                barrierItem.getChildren().add(cb);
                nodesMap.put(cb, n);
            }
            barrierItem.setExpanded(true);
        }

        tagTextField.setText("Command " + controller.getModel().nextCommandNumber());

        //initialize dialog
        dialog = new Dialog(parent, String.format("Add a command"));
        dialog.setMasthead("Add a command to the selected nodes");
        dialog.setGraphic(GlyphUtils.createDialogGlyph(FontAwesome.Glyph.PLUS, Color.GREEN));
        dialog.setContent(this);

        addCommandAction =
                new DialogAction("Add command", org.controlsfx.control.ButtonBar.ButtonType.OK_DONE, false, false, true, this::onAddCommandAction);


        dialog.getActions().setAll(addCommandAction, Dialog.ACTION_CANCEL);

        //initialize validation support
        validationSupport = new ValidationSupport();
        validationSupport.registerValidator(tagTextField, Validator.createEmptyValidator("Tag is a required field!"));
        validationSupport.registerValidator(commandTextField, Validator.createEmptyValidator("Command is a required field!"));
    }

    private void onAddCommandAction(ActionEvent actionEvent) {
        if (validationSupport.isInvalid()) {
            JFDialogs.create()
                    .message(String.format("Please correct the following issue(s): %s",
                            validationSupport.getValidationResult().getMessages().stream()
                                    .map(ValidationMessage::getText)
                                    .collect(Collectors.joining(", "))))
                    .showWarning();
            return;
        }

        boolean hasSelectedNode = false;

        for (CheckBoxTreeItem<String> cbti : nodesMap.keySet()) {
            if (cbti.isSelected()) {
                controller.addCommand(tagTextField.getText(), nodesMap.get(cbti).getUniqueId(),
                        barriersMap.get(cbti.getParent()).getOrderNumber()
                        , new Timestamp(startTextField.toMillis())
                        , new Timestamp(0)
                        , commandTextField.getText());
                hasSelectedNode = true;
            }
        }

        if (!hasSelectedNode) {
            JFDialogs.create()
                    .message("Select at least one node to which this command must be added!")
                    .showWarning();
            return;
        }

        dialog.setResult(addCommandAction);
        dialog.hide();
    }

    private AddMultipleCommandDialog(Window parent, ControllerController controller, List<FXRspecNode> nodes, ExperimentCommand c) {
        this(parent, controller, nodes);

        tagTextField.setText(c.getTag());
        startTextField.setTime(c.getStartingTime().toString());
        commandTextField.setText(c.getCommand());

        dialog.setTitle("Duplicate command");
    }

    private AddMultipleCommandDialog(Window parent, ControllerController controller, List<FXRspecNode> nodes, FXRspecNode n, ExperimentBarrierSegment b) {
        this(parent, controller, nodes);

        if (b != null) {
            barriersMap.keySet().stream()
                    .filter(cb -> barriersMap.get(cb).equals(b))
                    .forEach(cb -> {
                        //noinspection SuspiciousMethodCalls
                        cb.getChildren().stream()
                                .filter(nodeCB -> nodesMap.get(nodeCB).equals(n))
                                .forEach(nodeCB -> ((CheckBoxTreeItem<String>) nodeCB).setSelected(true));
                    });
        }

    }

    private AddMultipleCommandDialog(Window parent, ControllerController controller, List<FXRspecNode> nodes, FXRspecNode n, ExperimentBarrierSegment b, Timestamp t) {
        this(parent, controller, nodes, n, b);

        startTextField.setText(t.toString());
    }


    public void showDialog() {
        dialog.show();
    }

    public static void showSimpleAddDialog(Window parent, ControllerController controller, List<FXRspecNode> nodes) {
        AddMultipleCommandDialog npd = new AddMultipleCommandDialog(parent, controller, nodes);
        npd.showDialog();

    }

    public static void showSimpleAddDialog(Window parent, ControllerController controller, List<FXRspecNode> nodes, ExperimentCommand c) {
        AddMultipleCommandDialog npd = new AddMultipleCommandDialog(parent, controller, nodes, c);
        npd.showDialog();
    }

    public static void showSimpleAddDialog(Window parent, ControllerController controller, List<FXRspecNode> nodes, FXRspecNode n, ExperimentBarrierSegment b) {
        AddMultipleCommandDialog npd = new AddMultipleCommandDialog(parent, controller, nodes, n, b);
        npd.showDialog();
    }

    public static void showSimpleAddDialog(Window parent, ControllerController controller, List<FXRspecNode> nodes, FXRspecNode n, ExperimentBarrierSegment b, Timestamp t) {
        AddMultipleCommandDialog npd = new AddMultipleCommandDialog(parent, controller, nodes, n, b, t);

        npd.showDialog();
    }
}
