package be.iminds.ilabt.jfed.experimenter_gui.editor;

import be.iminds.ilabt.jfed.experimenter_gui.canvas.CanvasLink;
import be.iminds.ilabt.jfed.experimenter_gui.canvas.ExperimentCanvas;
import be.iminds.ilabt.jfed.experimenter_gui.canvas.impl.RspecCanvasLink;
import be.iminds.ilabt.jfed.experimenter_gui.canvas.impl.RspecCanvasNode;
import be.iminds.ilabt.jfed.highlevel.model.AuthorityList;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityFinder;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityListModel;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.util.GeniUrn;
import javafx.beans.property.BooleanProperty;
import javafx.beans.property.SimpleBooleanProperty;
import javafx.scene.CacheHint;
import javafx.scene.shape.Circle;
import javafx.scene.shape.Line;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * This class visualises a new canvas link in the making.
 * It is visualised by a line that starts from the supposed origin of  the link,
 * and ends with a circle when the target is unknown
 * <p/>
 * User: twalcari
 * Date: 11/8/13
 * Time: 5:47 PM
 */
public class NewCanvasLink {
    private final static Logger LOG = LoggerFactory.getLogger(NewCanvasLink.class);

    private final ExperimentCanvas experimentCanvas;
    private final AuthorityList authorityList;

    private final Line line;
    private final Circle circle;
    private final BooleanProperty visible = new SimpleBooleanProperty(false);
    private RspecCanvasNode originNode;
    private RspecCanvasLink originLink;
    private RspecCanvasNode targetNode;
    private RspecCanvasLink targetLink;


    public NewCanvasLink(ExperimentCanvas experimentCanvas, AuthorityList authorityList) {

        this.experimentCanvas = experimentCanvas;
        this.authorityList = authorityList;

        line = new Line();
        line.getStyleClass().add("new-canvas-link-line");
        line.setCache(true);
        line.setCacheHint(CacheHint.SPEED);
        line.visibleProperty().bind(visible);

        experimentCanvas.getCanvas().getChildren().add(0, line);

        circle = new Circle();
        circle.setRadius(3.5);
        circle.getStyleClass().add("new-canvas-link-end");
        circle.setCache(true);
        circle.setCacheHint(CacheHint.SPEED);
        circle.visibleProperty().bind(visible);

        experimentCanvas.getCanvas().getChildren().add(0, circle);


        //hidden by default
        setVisible(false);

    }

    public boolean isVisible() {
        return visible.get();
    }

    public void setVisible(boolean visible) {
        this.visible.set(visible);
    }

    public BooleanProperty visibleProperty() {
        return visible;
    }

    public RspecCanvasNode getOriginNode() {
        return originNode;
    }

    public void setOriginNode(RspecCanvasNode originNode) {
        this.originNode = originNode;
        if (originNode != null) {
            this.originLink = null;

            line.setStartX(CanvasLink.getCenterX(originNode));
            line.setStartY(CanvasLink.getCenterY(originNode));
        }
    }

    public RspecCanvasLink getOriginLink() {
        return originLink;
    }

    public void setOriginLink(RspecCanvasLink originLink) {
        this.originLink = originLink;
        if (originLink != null) {
            this.originNode = null;

            line.setStartX(originLink.getCenterX());
            line.setStartY(originLink.getCenterY());
        }
    }

    public void clear() {
        this.originLink = null;
        this.originNode = null;
        this.targetLink = null;
        this.targetNode = null;
    }

    public void setNewEnd(double x, double y) {
        line.setEndX(x);
        line.setEndY(y);

        circle.setCenterX(x);
        circle.setCenterY(y);
    }

    public RspecCanvasNode getTargetNode() {
        return targetNode;
    }

    public void setTargetNode(RspecCanvasNode node) {
        assert (originNode != null || originLink != null);

        this.targetNode = node;
        this.targetLink = null;

        setNewEnd(CanvasLink.getCenterX(node), CanvasLink.getCenterY(node));

    }

    public RspecCanvasLink getTargetLink() {
        return targetLink;
    }

    public void setTargetLink(RspecCanvasLink link) {
        assert (originNode != null || originLink != null);

        this.targetNode = null;
        this.targetLink = link;

        setNewEnd(link.getCenterX(), link.getCenterY());
    }

    /**
     * @return true if an origin and target are specified.
     */
    public boolean isLinkComplete() {
        return (originNode != null || originLink != null)
                && (targetNode != null || targetLink != null);
    }

    public boolean isLegalLink() {

        if (!isLinkComplete()) {
            LOG.trace("Requested link is not complete");
            return false;
        }

        if (originNode != null) {
            if (targetNode != null) {

                if (originNode == targetNode) {
                    LOG.trace("Origin and target-node cannot be identical.");
                }

                //do not allow self-links
                return originNode != targetNode;
            } else {
                assert (targetLink != null);

                //do not allow more than two interfaces on a stitched link
                if (targetLink.getRspecLink().isStitched(authorityList.getAuthorityListModel())) {
                    if (originNode == targetNode) {
                        LOG.trace("Cannot add node to a stitched link");
                    }
                    return false;
                }
                //prevent a normal link from becoming a stitched link
                if (isStitchedLink()) {
                    LOG.trace("Cannot add node from different AM to existing link");
                    return false;
                }

                //do not allow multiple interfaces to the same node to be added to a link

                if (targetLink.getRspecLink().getInterfaceForNode(originNode.getRspecNode()) != null) {
                    LOG.trace("Node already has an interface on the requested link.");
                    return false;
                }
                return true;

            }
        } else {
            assert (originLink != null);

            //do not allow more than two interface on a stitched link
            if (originLink.getRspecLink().isStitched(authorityList.getAuthorityListModel())) {
                LOG.trace("Cannot add node to a stitched link");
                return false;
            }

            if (targetNode != null) {
                //prevent a normal link from becoming a stitched link
                if (isStitchedLink()) {
                    LOG.trace("Cannot add node from different AM to existing link");
                    return false;
                }

                //do not allow multiple interfaces to the same node to be added to a link
                if (originLink.getRspecLink().getInterfaceForNode(targetNode.getRspecNode()) != null) {
                    LOG.trace("Node already has an interface on the requested link.");
                }

                return true;
            } else {
                //do not allow links between two RspecLinks
                LOG.trace("Cannot link two links together!");
                return false;
            }
        }
    }

    public boolean isStitchedLink() {
        if (!isLinkComplete())
            return false;

        GeniUrn originComponentManagerId = null;
        if (originNode != null)
            originComponentManagerId = originNode.getRspecNode().getComponentManagerId();
        else {
            assert originLink != null;
            assert originLink.getRspecLink().getInterfaces().size() > 0;
            originComponentManagerId =
                    originLink.getRspecLink().getInterfaces().get(0).getNode().getComponentManagerId();

            //check if link type is "lan" (required for stitching)
            if (originLink.getRspecLink().getLinkTypes().size() > 1 ||
                    (originLink.getRspecLink().getLinkTypes().size() == 1 &&
                            !(originLink.getRspecLink().getLinkTypes().get(0).equalsIgnoreCase("lan") || originLink.getRspecLink().getLinkTypes().get(0).equalsIgnoreCase("vlan")))
                    )
                return false;
        }

        assert originComponentManagerId != null;

        GeniUrn targetComponentManagerId = null;
        if (targetNode != null) {
            targetComponentManagerId = targetNode.getRspecNode().getComponentManagerId();
        } else {
            assert targetLink != null;
            assert targetLink.getRspecLink().getInterfaces().size() > 0;
            targetComponentManagerId =
                    targetLink.getRspecLink().getInterfaces().get(0).getNode().getComponentManagerId();

            //check if link type is "lan" (required for stitching)
            if (targetLink.getRspecLink().getLinkTypes().size() > 1 ||
                    (targetLink.getRspecLink().getLinkTypes().size() == 1 && !(targetLink.getRspecLink().getLinkTypes().get(0).equalsIgnoreCase("lan") || targetLink.getRspecLink().getLinkTypes().get(0).equalsIgnoreCase("vlan")))
                    )
                return false;
        }

        assert targetComponentManagerId != null;
        //return !originComponentManagerId.equals(targetComponentManagerId);

        //we need to check if the authority used in CreateSliver is the same. The URN might be different. See SfaAuthority.SpecialCase.PREFER_TOPLEVEL
        SfaAuthority originAuth = authorityList.finder().findByUrn(originComponentManagerId, AuthorityFinder.Purpose.CREATE_SLIVER);
        SfaAuthority targetAuth = authorityList.finder().findByUrn(targetComponentManagerId, AuthorityFinder.Purpose.CREATE_SLIVER);

        if (originAuth == null || targetAuth == null) {
            if (originAuth == null)
                LOG.warn("Could not find authority for " + originComponentManagerId + " -> falling back to comparing URNs");
            if (targetAuth == null)
                LOG.warn("Could not find authority for " + targetComponentManagerId + " -> falling back to comparing URNs");

            //fallback
            return !originComponentManagerId.equals(targetComponentManagerId);
        }

        return !originAuth.getUrn().equals(targetAuth.getUrn());
    }

    public boolean isOriginDefined() {
        return originNode != null || originLink != null;
    }
}
