package be.iminds.ilabt.jfed.experimenter_gui.editor.properties.cmi;

import be.iminds.ilabt.jfed.experimenter_gui.JFedConfiguration;
import be.iminds.ilabt.jfed.highlevel.controller.HighLevelController;
import be.iminds.ilabt.jfed.highlevel.controller.JFedHighLevelException;
import be.iminds.ilabt.jfed.highlevel.controller.TaskExecution;
import be.iminds.ilabt.jfed.highlevel.model.AuthorityInfo;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.lowlevel.fedmon.TestResult;
import be.iminds.ilabt.jfed.rspec.model.DiskImage;
import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec.model.SliverType;
import be.iminds.ilabt.jfed.rspec.model.SliverTypeBuilder;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXModelRspec;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecNode;
import be.iminds.ilabt.jfed.rspec.rspec_source.AdvertisementRspecSource;
import be.iminds.ilabt.jfed.util.GeniUrn;
import be.iminds.ilabt.jfed.util.ProgressHandler;
import com.google.common.collect.HashMultimap;
import com.google.common.collect.Multimap;
import javafx.beans.property.ObjectProperty;
import javafx.beans.property.SimpleObjectProperty;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import javafx.concurrent.Task;
import javafx.scene.control.ProgressBar;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.*;
import java.util.stream.Collectors;

/**
 * ComponentManagerInfo contains info about what a particular component manager has to offer.
 * It combines local information (from the testbed configuration-xml's), with information gathered from listResources-calls.
 */
public class ComponentManagerInfo {
    private static final Logger LOG = LoggerFactory.getLogger(ComponentManagerInfo.class);
    private static final double PROGRESS_FETCH_REQUESTED = ProgressBar.INDETERMINATE_PROGRESS;
    private static final double PROGRESS_FETCH_COMPLETE = 0.0;
    private static final double PROGRESS_PARSE_COMPLETE = 0.95;

    private final HighLevelController hlc;


    private final AuthorityInfo authorityInfo;

    private final SliverType defaultSliverType;
    private final ObservableList<SliverType> sliverTypes = FXCollections.observableArrayList();
    private final ObservableList<DiskImage> diskImages = FXCollections.observableArrayList();

    private final boolean requiresFixedNodeAssignment;
    private final boolean linkSupport;

    private final AvailableResources availableResources;
    private final AllResources allResources;

    private Map<GeniUrn, TestResult> fedmonResults;

    public ComponentManagerInfo(AuthorityInfo authorityInfo, @Nullable JFedConfiguration.TestbedDescription testbedDescription, HighLevelController hlc) {
        this.hlc = hlc;
        assert testbedDescription == null || authorityInfo.getSfaAuthority().getUrn().equals(testbedDescription.getUrn())
                : "Mismatch between authorityInfo and testbedDescription!";

        SliverType defaultSliverType = null;
        this.authorityInfo = authorityInfo;
        if (testbedDescription != null) {
            for (JFedConfiguration.SliverType st : testbedDescription.getSliverTypes()) {
                SliverTypeBuilder stb = new SliverTypeBuilder(st.getName());
                if (st.getDiskImageName() != null)
                    stb.addDiskImage(st.getDiskImageName());
                SliverType storedSt = stb.build();
                if (defaultSliverType == null)
                    defaultSliverType = storedSt;
                sliverTypes.add(storedSt);
            }
            requiresFixedNodeAssignment = testbedDescription.requiresFixedNodeAssignment();
            linkSupport = testbedDescription.hasLinkSupport();
        } else {
            //set default values
            sliverTypes.add(new SliverTypeBuilder("raw-pc").build());
            requiresFixedNodeAssignment = false;
            linkSupport = true;
        }

        this.defaultSliverType = defaultSliverType;

        this.availableResources = new AvailableResources();
        this.allResources = new AllResources();

    }

    /**
     * Returns non-node specific sliverTypes (e.g. all available sliverTypes on the componentManager)
     *
     * @return list of sliver-types
     */
    public ObservableList<SliverType> getSliverTypes() {
        return FXCollections.unmodifiableObservableList(sliverTypes);
    }

    /**
     * get the first sliver type mentioned in the config. This one should always be chosen when an arbitrary choice needs to be made.
     *
     * @return the default sliver type
     */
    @Nullable
    public SliverType getDefaultSliverType() {
        return defaultSliverType;
    }

    public AuthorityInfo getAuthorityInfo() {
        return authorityInfo;
    }

    public SfaAuthority getAuthority() {
        return authorityInfo.getSfaAuthority();
    }

    public ObservableList<DiskImage> getDiskImages() {
        return diskImages;
    }


    public boolean hasSliverType(@Nonnull String sliverTypeName) {
        return sliverTypes.stream()
                .filter(type -> sliverTypeName.equals(type.getName()))
                .findAny().isPresent();
    }

    public boolean isRequiresFixedNodeAssignment() {
        return requiresFixedNodeAssignment;
    }

    public boolean isLinkSupport() {
        return linkSupport;
    }

    public void updateDiskImagesAndSliverTypes() {

        AdvertisementRspecSource advertisementRspecSource = allResources.getAdvertisementRspecSource();
        if (advertisementRspecSource == null)
            advertisementRspecSource = availableResources.getAdvertisementRspecSource();

        assert advertisementRspecSource != null;

        FXModelRspec modelRspec = (FXModelRspec) advertisementRspecSource.getModelRspec(ModelRspecType.FX);


        //find all available sliver-types and disk-images
        Multimap<String, DiskImage> newSliverTypesMap = HashMultimap.create();
        Set<String> newSliverTypesWithoutDiskImages = new HashSet<>();

        Set<DiskImage> newDiskImages = new HashSet<>();
        Multimap<DiskImage, String> sliverTypesPerDiskImage = HashMultimap.create();
        for (FXRspecNode node : modelRspec.getNodes()) {
            for (SliverType nodeSliverType : node.getSliverTypes()) {
                if (!nodeSliverType.getDiskImages().isEmpty()) {

                    for (DiskImage diskImage : nodeSliverType.getDiskImages()) {
                        newSliverTypesMap.put(nodeSliverType.getName(), diskImage);
                        sliverTypesPerDiskImage.put(diskImage, nodeSliverType.getName());
                    }
                    newDiskImages.addAll(nodeSliverType.getDiskImages());
                } else {
                    newSliverTypesWithoutDiskImages.add(nodeSliverType.getName());
                }
            }
        }


        //collect all sliver-types mentioned in the manifest: first the ones with disk-images ...
        List<SliverType> newSliverTypesList =
                newSliverTypesMap.asMap().entrySet().stream()
                        .map(entry -> new SliverTypeBuilder(entry.getKey()).setDiskImages(entry.getValue()).build())
                        .collect(Collectors.toList());

        // ... and then the ones without!
        newSliverTypesList.addAll(
                newSliverTypesWithoutDiskImages.stream()
                        .map(sliverTypeName -> new SliverTypeBuilder(sliverTypeName).build())
                        .collect(Collectors.toList()));

        sliverTypes.setAll(newSliverTypesList);
        sliverTypes.sort(Comparator.comparing(SliverType::getName));


        diskImages.setAll(newDiskImages);
        diskImages.sort(
                Comparator.comparing(DiskImage::getDescription, Comparator.nullsLast(Comparator.<String>naturalOrder()))
                        .thenComparing(DiskImage::getName, Comparator.nullsLast(Comparator.<String>naturalOrder()))
                        .thenComparingLong(DiskImage::hashCode)
        );
    }

    public AvailableResources getAvailableResources() {
        return availableResources;
    }

    public AllResources getAllResources() {
        return allResources;
    }


    public class AvailableResources extends Resources {

        @Override
        public HighLevelController.GetAdvertisementTask getAdvertisementTask() {
            return hlc.getAdvertisement(getAuthority(), true /*only available */);
        }

        @Override
        public HighLevelController.FedmonResultsTask getFedmonResultsTask() {
            return hlc.getFedmonResultsTask("ple", "autonodelogin", false /*successOnly*/); //do not filter out failures! We need them.
        }

        @Override
        public AdvertisementRspecSource getAdvertisementRspecSource() {
            return authorityInfo.getAvailableAdvertisementRspec();
        }

        @Override
        public boolean mustRemoveUnavailable() {
            return true;
        }
    }


    public class AllResources extends Resources {

        @Override
        public HighLevelController.GetAdvertisementTask getAdvertisementTask() {
            return hlc.getAdvertisement(getAuthority(), false /*only available */);
        }

        @Override
        public HighLevelController.FedmonResultsTask getFedmonResultsTask() {
            return hlc.getFedmonResultsTask("ple", "autonodelogin", false /*successOnly */);
        }

        @Override
        public AdvertisementRspecSource getAdvertisementRspecSource() {
            return authorityInfo.getAllAdvertisementRspec();
        }

        @Override
        public boolean mustRemoveUnavailable() {
            return false;
        }
    }


    public abstract class Resources {

        private final ObservableList<FXRspecNode> advertisementComponentIds = FXCollections.observableArrayList();

        private final ObjectProperty<RefreshAdvertisementTask> currentRefreshAdvertisementTask = new SimpleObjectProperty<>();

        public abstract HighLevelController.FedmonResultsTask getFedmonResultsTask();
        public abstract HighLevelController.GetAdvertisementTask getAdvertisementTask();

        public abstract AdvertisementRspecSource getAdvertisementRspecSource();

        /** if true, nodes reported as unavailable by FedMon will be removed. */
        public abstract boolean mustRemoveUnavailable();

        public ObservableList<FXRspecNode> getAdvertisementComponentIds() {
            return advertisementComponentIds;
        }

        /**
         * Checks if the "available component-ids" for this component manager are available
         */
        public boolean isComponentIdsAvailable() {
            return getAdvertisementRspecSource() != null;
        }


        /**
         * Starts the refresh of the available componentIds in Component Manager.
         * This refresh is performed asynchronously, and a task is returned which allows you to follow the progress.
         *
         * @return The task that performs the action. Can be used by a GUI to track progress
         */
        public RefreshAdvertisementTask refreshAdvertisement() {
            if (currentRefreshAdvertisementTask.get() == null) {
                currentRefreshAdvertisementTask.set(new RefreshAdvertisementTask());
                new Thread(currentRefreshAdvertisementTask.get()).start();
            }

            return currentRefreshAdvertisementTask.get();
        }

        public class RefreshAdvertisementTask extends Task<Void> {
            @Override
            protected Void call() throws Exception {
                HighLevelController.GetAdvertisementTask fetchAdvertisementTask = getAdvertisementTask();

                TaskExecution taskExecution = hlc.submitTask(fetchAdvertisementTask);
                updateProgress(PROGRESS_FETCH_REQUESTED, 1);

                TaskExecution taskExecution2 = null;
                HighLevelController.FedmonResultsTask fedmonNodeStatusTask = null;
                if (getAuthority().getUrn().equals(GeniUrn.parse("urn:publicid:IDN+ple:ibbtple+authority+cm"))) {
                    LOG.debug("Fetching fedmon results in addition to advertisement.");
                    fedmonNodeStatusTask = getFedmonResultsTask();

                    taskExecution2 = hlc.submitTask(fedmonNodeStatusTask);
                }

                while ((!taskExecution.isCompleted()) || (taskExecution2 != null && !taskExecution2.isCompleted())) {
                    Thread.sleep(100);
                }
                assert taskExecution.isCompleted();
                assert taskExecution2 == null || taskExecution2.isCompleted();

                if (taskExecution.getState() != TaskExecution.TaskState.SUCCESS) {
                    throw new JFedHighLevelException("Unexpected state for getAdvertisement task: " + taskExecution.getState(),
                            taskExecution.getException());
                }

                //HACK wait for availableAdvertismentRspec to become available, this happens by a process in another thread..

                while (getAdvertisementRspecSource() == null) {
                    Thread.sleep(100);
                }

                updateProgress(PROGRESS_FETCH_COMPLETE, 1);

                assert getAdvertisementRspecSource() != null;

                ProgressHandler parserProgressHandler =
                        (progress, totalWork) -> updateProgress(PROGRESS_FETCH_COMPLETE +
                                        (PROGRESS_PARSE_COMPLETE - PROGRESS_FETCH_COMPLETE) * (progress / totalWork),
                                1);

                //register parserProgressHandler
                getAdvertisementRspecSource().getModelRspec(parserProgressHandler);

                if (fedmonNodeStatusTask != null) {
                    if (taskExecution2.getState() != TaskExecution.TaskState.SUCCESS)
                        LOG.warn("Fedmon Status task status is "+taskExecution2.getState());

                    List<TestResult> testResults = fedmonNodeStatusTask.getResults();
                    if (fedmonResults == null)
                        fedmonResults = new HashMap<GeniUrn, TestResult>();

                    if (testResults != null)
                        if (!testResults.isEmpty())
                            for (TestResult testResult : testResults)
                                if (testResult.getFixedNodeUrn() != null) {
                                    GeniUrn nodeUrn = GeniUrn.parse(testResult.getFixedNodeUrn());
                                    if (nodeUrn != null)
                                        fedmonResults.put(nodeUrn, testResult);
                                    else
                                        LOG.warn("Invalid node urn in fedmon reply: \"" + testResult.getFixedNodeUrn() + "\"");
                                } else
                                    LOG.warn("Node without urn in fedmon reply id=" + testResult.getId());
                        else
                            LOG.warn("Fedmon Status task has empty results");
                    else
                        LOG.warn("Fedmon Status task has null results");
                }

                updateProgress(PROGRESS_PARSE_COMPLETE, 1);

                return null;
            }

            @Override
            protected void scheduled() {
                updateProgress(0, 1);
            }

            @Override
            protected void succeeded() {
                updateComponentIds();

                //remove the current refresh task from the ComponentManagerInfo
                currentRefreshAdvertisementTask.set(null);
            }
        }

        public RefreshAdvertisementTask getCurrentRefreshAdvertisementTask() {
            return currentRefreshAdvertisementTask.get();
        }

        public ObjectProperty<RefreshAdvertisementTask> currentRefreshAdvertisementTaskProperty() {
            return currentRefreshAdvertisementTask;
        }

        /**
         * Returns a list with available component-Ids.
         * <p/>
         * This method will only passivly fill the CompomentManagerInfo-object if there is
         * information available from a ListResources-call. It will not actively perform one (use refreshAdvertisement for that)!
         */
        public void updateComponentIds() {
            //check if we have an advertisement Rspec available. If not, we have nothing to populate!
            if (getAdvertisementRspecSource() == null
                    || getAdvertisementRspecSource().getModelRspec(ModelRspecType.FX) == null) {
                LOG.trace("No detailed node-info available for testbed {}.", getAuthority().getName());
                return;
            }

            FXModelRspec modelRspec =
                    (FXModelRspec) getAdvertisementRspecSource().getModelRspec(ModelRspecType.FX);

            //make a copy, so that it can be sorted!
            List<FXRspecNode> nodes = new ArrayList<>(modelRspec.getNodes());
            for (Iterator<FXRspecNode> it = nodes.listIterator(); it.hasNext(); ) {
                FXRspecNode node = it.next();
                if (node == null || node.getComponentId() == null) {
                    LOG.warn("Item in received node list has no component_id. Will be ignored!");
                    it.remove();
                    continue;
                }
                //use component_id as fallback when component_name is not present
                if (node.getComponentName() == null)
                    node.setComponentName(node.getComponentId().toString());

                //FEDIBBTDEV-1591 remove unavailble nodes that are reported in an "available" advertisement RSpec (= workaround for AM bug)
                if (node.getAvailable() == Boolean.FALSE && mustRemoveUnavailable()) { //dirty
                    it.remove();
                    continue;
                }

                //if there is information from fedmon, use it to update the availability of the node
                if (fedmonResults != null) {
                    TestResult testResult = fedmonResults.get(node.getComponentId());
                    if (testResult != null) {
                        node.setAvailable(testResult.isSuccessSummary());
                        if (!testResult.isSuccessSummary() && mustRemoveUnavailable()) { //dirty
                            it.remove();
                            continue;
                        }
                    }
                }
            }

            //noinspection ConstantConditions
            nodes.sort(Comparator.comparing(FXRspecNode::getComponentName));

            LOG.debug("Added {} nodes to the list of testbed {}",
                    modelRspec.getNodes().size(), getAuthority().getName());

            //sort them according to their name

            advertisementComponentIds.setAll(nodes);

            updateDiskImagesAndSliverTypes();

        }


    }

}
