package be.iminds.ilabt.jfed.experimenter_gui.preferences.subsections;

import be.iminds.ilabt.jfed.experimenter_gui.preferences.AbstractPreferencesSubPane;
import be.iminds.ilabt.jfed.experimenter_gui.util.ui.TextFieldWithStatus;
import be.iminds.ilabt.jfed.lowlevel.userloginmodel.UserLoginModelManager;
import be.iminds.ilabt.jfed.preferences.GuiPreferenceKey;
import be.iminds.ilabt.jfed.preferences.JFedGuiPreferences;
import be.iminds.ilabt.jfed.ssh_terminal_tool.putty.PuTTYPrivateKeyFile;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import javafx.event.ActionEvent;
import javafx.fxml.FXML;
import javafx.scene.Node;
import javafx.scene.control.Button;
import javafx.scene.control.RadioButton;
import javafx.stage.FileChooser;

import java.io.File;
import java.io.IOException;

public abstract class SshAuthenticationPane extends AbstractPreferencesSubPane {
    @FXML
    protected Node root;
    @FXML
    protected RadioButton useCertificateRadioButton;
    @FXML
    protected RadioButton useCustomKeyPairRadioButton;
    @FXML
    protected TextFieldWithStatus fileTextField;
    @FXML
    protected Button browseButton;

    private final UserLoginModelManager userLoginModelManager;
    protected final JFedGuiPreferences jFedPreferences;

    SshAuthenticationPane(UserLoginModelManager userLoginModelManager, JFedGuiPreferences jFedPreferences) {
        super("SSH Authentication", false);
        this.userLoginModelManager = userLoginModelManager;
        this.jFedPreferences = jFedPreferences;
    }

    @FXML
    protected void initialize() {
        fileTextField.disableProperty().bind(useCustomKeyPairRadioButton.selectedProperty().not());
        browseButton.disableProperty().bind(useCustomKeyPairRadioButton.selectedProperty().not());

        fileTextField.focusedProperty().addListener((observableValue, oldValue, newValue) -> {
            if (newValue) {
                fileTextField.setStatus(TextFieldWithStatus.Status.NONE);
            } else {
                testPrivateKey();
            }
        });

        if (jFedPreferences.getBoolean(GuiPreferenceKey.PREF_SSH_OVERRIDE)) {
            useCustomKeyPairRadioButton.setSelected(true);

            if (jFedPreferences.getFile(GuiPreferenceKey.PREF_SSH_PUTTY_KEY_FILE) != null) {
                fileTextField.setText(jFedPreferences.getFile(GuiPreferenceKey.PREF_SSH_PUTTY_KEY_FILE).getPath());
            } else {
                if (jFedPreferences.getFile(GuiPreferenceKey.PREF_SSH_PRIVATE_KEY_FILE) != null) {
                    fileTextField.setText(jFedPreferences.getFile(GuiPreferenceKey.PREF_SSH_PRIVATE_KEY_FILE).getPath());
                }
            }
        } else {
            useCustomKeyPairRadioButton.setSelected(false);
            fileTextField.setText("");
            fileTextField.setStatus(TextFieldWithStatus.Status.NONE);
        }

        //evaluate inputs
        if (useCustomKeyPairRadioButton.isSelected())
            testPrivateKey();
    }

    protected void testPrivateKey() {
        if (fileTextField.getText().length() == 0) {
            fileTextField.setStatus(TextFieldWithStatus.Status.NONE);
            return;
        }

        try {
            if (testPrivateKeyFile(new File(fileTextField.getText()))) {
                fileTextField.setStatus(TextFieldWithStatus.Status.OK);

            } else {
                JFDialogs.create()
                        .owner(root)
                        .message(" Please select a valid private key.")
                        .masthead("The provided key is not valid.")
                        .title("Invalid key file")
                        .showWarning();
                fileTextField.setStatus(TextFieldWithStatus.Status.ERROR);
            }
        } catch (IOException ex) {
            JFDialogs.create()
                    .owner(root)
                    .message("An error occured while processing the key file")
                    .masthead("Error")
                    .title("Error")
                    .showException(ex);

            fileTextField.setStatus(TextFieldWithStatus.Status.ERROR);
        }
    }

    @FXML
    protected void onBrowseButtonAction(ActionEvent actionEvent) {
        FileChooser fc = new FileChooser();
        fc.setTitle("Select your Private Key");
        if (getLoggedInUserPrivateKeyFile() != null)
            fc.setInitialDirectory(getLoggedInUserPrivateKeyFile().getParentFile());

        File selectedFile = fc.showOpenDialog(fileTextField.getScene().getWindow());
        if (selectedFile != null) {
            fileTextField.setText(selectedFile.getAbsolutePath());
            testPrivateKey();
        }

    }

    protected abstract boolean testPrivateKeyFile(File file) throws IOException;

    protected File getLoggedInUserPrivateKeyFile() {
        File privateKey;

        switch (userLoginModelManager.getUserLoginModelType()) {
            case KEY_CERT_INTERNAL_INFO:
            case KEY_CERT_EXTERNAL_INFO:
                privateKey = userLoginModelManager.getKeyCertUserLoginModel().getKeyCertFile();
                break;
            case PLANETLAB:
                privateKey = userLoginModelManager.getPlanetlabUserLoginModel().getSshPrivateKeyFile();
                break;
            default:
                throw new RuntimeException("Unexpected UserLoginModelType");
        }
        return privateKey;
    }

    @Override
    public boolean check() {
        if (useCustomKeyPairRadioButton.isSelected()) {
            if (fileTextField.getStatus() != TextFieldWithStatus.Status.OK) {
                JFDialogs.create()
                        .owner(root)
                        .message("The provided private key file is not valid")
                        .showError();
                return false;
            }

            File newPrivKeyFile = new File(fileTextField.getText());
            try {
                PuTTYPrivateKeyFile.isPuttyPrivateKey(newPrivKeyFile);
            } catch (IOException e) {
                JFDialogs.create()
                        .owner(root)
                        .message("The provided private key file is not valid")
                        .showError();
                return false;
            }
        }

        return true;
    }

    @Override
    public boolean save() {
        if (useCustomKeyPairRadioButton.isSelected()) {
            if (fileTextField.getStatus() != TextFieldWithStatus.Status.OK) {
                JFDialogs.create()
                        .owner(root)
                        .message("The provided private key file is not valid")
                        .showError();
                return false;
            }

            File newPrivKeyFile = new File(fileTextField.getText());
            boolean isPuttyFile;
            try {
                isPuttyFile = PuTTYPrivateKeyFile.isPuttyPrivateKey(newPrivKeyFile);
            } catch (IOException e) {
                JFDialogs.create()
                        .owner(root)
                        .message("The provided private key file is not valid")
                        .showError();
                return false;
            }
            if (isPuttyFile) {
                jFedPreferences.clear(GuiPreferenceKey.PREF_SSH_PRIVATE_KEY_FILE);
                jFedPreferences.setFile(GuiPreferenceKey.PREF_SSH_PUTTY_KEY_FILE, newPrivKeyFile);
            } else {
                jFedPreferences.clear(GuiPreferenceKey.PREF_SSH_PUTTY_KEY_FILE);
                jFedPreferences.setFile(GuiPreferenceKey.PREF_SSH_PRIVATE_KEY_FILE, newPrivKeyFile);
            }

            jFedPreferences.setBoolean(GuiPreferenceKey.PREF_SSH_OVERRIDE, true);
        } else
            jFedPreferences.setBoolean(GuiPreferenceKey.PREF_SSH_OVERRIDE, false);

        return true;
    }
}
