package be.iminds.ilabt.jfed.experimenter_gui.slice;

import be.iminds.ilabt.jfed.experimenter_gui.slice.events.RenewExperimentEvent;
import be.iminds.ilabt.jfed.experimenter_gui.util.ExecuteOnNotNull;
import be.iminds.ilabt.jfed.highlevel.model.Slice;
import be.iminds.ilabt.jfed.highlevel.model.Sliver;
import be.iminds.ilabt.jfed.preferences.JFedGuiPreferences;
import com.google.common.eventbus.EventBus;
import com.google.common.util.concurrent.ThreadFactoryBuilder;
import javafx.application.Platform;
import javafx.beans.WeakInvalidationListener;
import javafx.collections.ListChangeListener;
import javafx.util.Duration;
import org.controlsfx.control.Notifications;
import org.controlsfx.control.action.Action;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;

/**
 * User: twalcari
 * Date: 8/13/2015
 * Time: 17:17
 */
public class ExpirationNotifier {
    private static final Logger LOG = LoggerFactory.getLogger(ExpirationNotifier.class);
    private static final ScheduledExecutorService executor =
            Executors.newSingleThreadScheduledExecutor(new ThreadFactoryBuilder().setNameFormat("ExpirationNotifier-%d").build());

    private final JFedGuiPreferences jFedPreferences;
    private final EventBus eventBus;

    private final ExperimentViewController experimentViewController;
    private final Experiment experiment;

    private final WeakInvalidationListener listener =
            new WeakInvalidationListener(observable -> onExpirationTimesUpdated());

    private ScheduledFuture<?> scheduledFuture;

    private boolean warningShown = false;
    private NextExperimentExpiration nextExperimentExpiration;

    public ExpirationNotifier(ExperimentViewController experimentViewController, JFedGuiPreferences jFedPreferences, EventBus eventBus) {
        this.experimentViewController = experimentViewController;
        this.experiment = experimentViewController.getExperiment();
        this.jFedPreferences = jFedPreferences;
        this.eventBus = eventBus;

        new ExecuteOnNotNull<Slice>(experiment.sliceProperty()) {
            @Override
            public void run() {
                Slice slice = experiment.getSlice();

                slice.expirationDateProperty().addListener(listener);
                slice.sliversProperty().addListener((ListChangeListener<Sliver>) c -> {
                    while (c.next()) {
                        if (c.wasAdded()) {
                            c.getAddedSubList().forEach(sliver -> sliver.expirationDateProperty().addListener(listener));
                        }
                    }
                });

                slice.getSlivers().forEach(sliver -> sliver.expirationDateProperty().addListener(listener));

                onExpirationTimesUpdated();
            }
        };

        scheduledFuture = executor.scheduleWithFixedDelay(this::checkToShowWarning, 0, 30, TimeUnit.SECONDS);
    }

    private void checkToShowWarning() {
        if (warningShown) {
            LOG.trace("Expiration warning for {} already shown. Ignoring.", experiment.getName());
            return;
        }

        if (experiment.getSliceState() != SliceState.TESTING && experiment.getSliceState() != SliceState.READY) {
            LOG.debug("Not checking for expiration as experiment is currently {}", experiment.getSliceState());
            return;
        }

        nextExperimentExpiration = new NextExperimentExpiration(experiment.getSlice());

        if (nextExperimentExpiration.getFirstExpirationTime() == null) {
            LOG.info("Could not find a valid expiration time to compare against for experiment {}", experiment.getName());
            return;
        }


        long minutesUntilExpiration =
                Instant.now().until(nextExperimentExpiration.getFirstExpirationTime(), ChronoUnit.MINUTES);
        if (minutesUntilExpiration < (long) (jFedPreferences.getExpirationWarningTime())) {
            showExpirationWarning();
        } else {
            LOG.trace("Experiment {} is expiring in {} minutes, which is longer then set expiration warning time {}.",
                    experiment.getName(),
                    minutesUntilExpiration,
                    jFedPreferences.getExpirationWarningTime());
        }
    }

    private void onExpirationTimesUpdated() {
        warningShown = false;
    }

    private void showExpirationWarning() {

        long fullMinutesUntilExpiration =
                Instant.now().until(nextExperimentExpiration.getFirstExpirationTime(), ChronoUnit.MINUTES);

        if (jFedPreferences.isExpirationWarningEnabled() && fullMinutesUntilExpiration >= 0) {

            StringBuilder message = new StringBuilder();

            if (nextExperimentExpiration.getFirstExpiringSliver() == null) {
                message.append("The experiment '").append(experiment.getName()).append("' is");
            } else {
                message.append("The resources on ").append(nextExperimentExpiration.getFirstExpiringSliver().getAuthority().getHrn())
                        .append(" of experiment '").append(experiment.getName()).append("' are");
            }


            message.append(" expiring in less then ").append(fullMinutesUntilExpiration + 1).append(" minutes.");


            Action renewAction = new Action("Renew now",
                    event -> eventBus.post(new RenewExperimentEvent(experimentViewController.getScene().getWindow(),
                            experimentViewController.getExperimentController())));


            LOG.info("Showing expiration warning for experiment {} to user.", experiment.getName());

            Platform.runLater(() -> {
                Notifications.create()
                        .title(nextExperimentExpiration.getFirstExpiringSliver() == null
                                ? "Experiment expiring!" : "Resources expiring!")
                        .text(message.toString())
                        .action(renewAction)
                        .hideAfter(Duration.millis(java.time.Duration.between(Instant.now(), nextExperimentExpiration.getFirstExpirationTime()).toMillis()))
                        .showWarning();
            });


        } else {
            LOG.info("Not showing expiration warning, as user disabled it!");
        }

        warningShown = true;


    }

    @Override
    protected void finalize() throws Throwable {
        super.finalize();
        scheduledFuture.cancel(true);
    }
}
