package be.iminds.ilabt.jfed.experimenter_gui.util;

import be.iminds.ilabt.jfed.experimenter_gui.preferences.PreferencesDialogFactory;
import be.iminds.ilabt.jfed.highlevel.model.Slice;
import be.iminds.ilabt.jfed.lowlevel.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnection;
import be.iminds.ilabt.jfed.rspec.model.BasicStringRspec;
import be.iminds.ilabt.jfed.rspec.model.LoginService;
import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXModelRspec;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecNode;
import be.iminds.ilabt.jfed.rspec.rspec_source.ManifestRspecSource;
import be.iminds.ilabt.jfed.ssh_terminal_tool.ExternalSshTerminal;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import com.google.common.base.Preconditions;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.inject.Inject;
import javax.inject.Singleton;
import java.util.List;
import java.util.Optional;

/**
 * User: twalcari
 * Date: 1/29/2015
 * Time: 16:13
 */
@Singleton
public class SSHTerminalUtil {
    private static final Logger LOG = LoggerFactory.getLogger(SSHTerminalUtil.class);


    private final ExternalSshTerminal externalSshTerminal;
    private final PreferencesDialogFactory preferencesDialogFactory;
    private final GeniUserProvider geniUserProvider;
    private final ProxyServiceUtil proxyServiceUtil;


    @Inject
    public SSHTerminalUtil(ExternalSshTerminal externalSshTerminal, PreferencesDialogFactory preferencesDialogFactory,
                           GeniUserProvider geniUserProvider, ProxyServiceUtil proxyServiceUtil) {
        this.externalSshTerminal = externalSshTerminal;
        this.preferencesDialogFactory = preferencesDialogFactory;
        this.geniUserProvider = geniUserProvider;
        this.proxyServiceUtil = proxyServiceUtil;
    }

    public void launchSSHTerminalForcedLocalUser(FXRspecNode rspecNode, ManifestRspecSource manifestRspecSource) {
        String username =
                geniUserProvider.getLoggedInGeniUser().getUserUrn().getEncodedResourceName();
        launchSSHTerminal(rspecNode, manifestRspecSource, username);
    }


    public void launchSSHTerminalForcedLocalUser(FXRspecNode rspecNode, Slice slice) {
        String username =
                geniUserProvider.getLoggedInGeniUser().getUserUrn().getEncodedResourceName();
        launchSSHTerminal(rspecNode, slice, username);
    }

    public void launchSSHTerminal(FXRspecNode rspecNode, Slice slice) {
        launchSSHTerminal(rspecNode, slice, null);
    }

    public void launchSSHTerminal(@Nonnull FXRspecNode rspecNode, @Nullable Slice slice, @Nullable String overrideUserName) {
        //get the associated manifest RSpec-node
        if (slice == null) {
            LOG.warn("Could not open SSH terminal: manifestRspecSource=null");
            JFDialogs.create()
                    .message("There is currently insufficient information available to login.")
                    .showError();
        } else {
            launchSSHTerminal(rspecNode, slice.getManifestRspec(), overrideUserName);
        }
    }


    public void launchSSHTerminal(FXRspecNode rspecNode, ManifestRspecSource manifestRspecSource) {
        launchSSHTerminal(rspecNode, manifestRspecSource, null);
    }

    /**
     * The rspecNode is NOT necessarily a part of the provided manifestRspecSource. It can be that the rspecNode is from the RequestRspec.
     * <p>
     * We first have to match the rspecNode to the corresponding node in the manifstRspecSource!
     */
    public void launchSSHTerminal(@Nonnull FXRspecNode rspecNode, @Nullable ManifestRspecSource manifestRspecSource,
                                  @Nullable String overrideUsername) {
        //get the associated manifest RSpec-node
        if (manifestRspecSource == null) {
            LOG.warn("launchSSHTerminal: Could not open SSH terminal: manifestRspecSource=null");
            JFDialogs.create()
                    .message("There is currently insufficient information available to login.")
                    .showError();
            return;
        }

        LOG.warn("launchSSHTerminal called with rspecNode "+rspecNode.getUniqueId()+" overrideUsername="+overrideUsername);

        //first try to fetch the loginservices from the modelRspec
        FXModelRspec modelRspec = (FXModelRspec) manifestRspecSource.getModelRspec(ModelRspecType.FX);

        if (modelRspec != null) {
            FXRspecNode manifestNode = modelRspec.getNodeByUniqueId(rspecNode.getUniqueId());

            if (manifestNode == null) {
                LOG.warn("launchSSHTerminal: Did not find node by uniqueID "+rspecNode.getUniqueId()+" will try with clientId "+rspecNode.getClientId());
                manifestNode = modelRspec.getNodeByClientId(rspecNode.getClientId());
            }

            if (manifestNode != null) {
                LOG.debug("launchSSHTerminal: Found node in manifest model");
//                final List<LoginService> loginServices = manifestNode.getLoginServices();
                final List<LoginService> loginServices = proxyServiceUtil.nonProxyLoginServices(manifestNode);

                //try to find a loginService that matches the desired username

                String desiredUsername = overrideUsername == null ?
                        geniUserProvider.getLoggedInGeniUser().getUserUrn().getEncodedResourceName() :
                        overrideUsername;

                Optional<LoginService> loginService =
                        loginServices.stream().filter(ls -> ls.getUsername().equals(desiredUsername)).findFirst();

                if (!loginService.isPresent() && !loginServices.isEmpty()) {
                    LOG.debug("launchSSHTerminal: No login service with username "+desiredUsername+". Trying fallback to other user.");
                    if (loginServices.isEmpty()) {
                        LOG.debug("launchSSHTerminal: No non proxy login service . Trying fallback to proxy login service.");
                        if (manifestNode.getLoginServices().isEmpty())
                            loginService = Optional.empty();
                        else
                            loginService = Optional.of(manifestNode.getLoginServices().get(0));
                    }
                    else
                        loginService = Optional.of(loginServices.get(0));
                }

                if (loginService.isPresent()) {
                    LOG.debug("launchSSHTerminal: found login service for user "+loginService.get().getUsername());
                    JFedConnection.SshProxyInfo testbedProxyInfo = proxyServiceUtil.findTestbedProxy(manifestNode, loginService.get());
                    if (testbedProxyInfo == null)
                        LOG.debug("launchSSHTerminal: Did not find testbed proxy info for login "+
                                loginService.get().getUsername()+"@"+loginService.get().getHostname()+":"+loginService.get().getPort());
                    else
                        LOG.debug("launchSSHTerminal: Found testbed proxy info: "+testbedProxyInfo);
                    launchSSHTerminal(loginService.get(), testbedProxyInfo, overrideUsername);
                    return;
                } else
                    LOG.debug("launchSSHTerminal: no login service found in manifest model node");
            } else
                LOG.warn("launchSSHTerminal: Could not find node in manifest model.");
        }

        LOG.warn("launchSSHTerminal: Could not find loginService via modelRspec, doing fallback.");


        final List<BasicStringRspec.LoginService> loginServices =
                manifestRspecSource.getNodeLoginInfo(rspecNode);


        if (loginServices == null || loginServices.isEmpty()) {
            LOG.warn("launchSSHTerminal: There are no loginServices available for {}", rspecNode.getClientId());
            JFDialogs.create()
                    .message("The selected node doesn't provide any information to perform an interactive login.")
                    .masthead("Could not open SSH terminal")
                    .title("Could not open SSH terminal")
                    .showException(new RuntimeException(
                            "node id=\"" + rspecNode.getUniqueId() + "\" "
                                    + manifestRspecSource.getStringRspec().getXmlString()
                    ));

            return;
        }

        //try to find a loginService with a login-name equal to that of the logged-in user
        GeniUser geniUser = geniUserProvider.getLoggedInGeniUser();

        BasicStringRspec.LoginService chosenLoginService = loginServices.get(0);
        for (int i = 1; i < loginServices.size(); i++) {
            BasicStringRspec.LoginService otherLoginService = loginServices.get(i);
            if (otherLoginService.getUsername().equals(geniUser.getUserUrn().getEncodedResourceName()))
                chosenLoginService = otherLoginService;
        }

        launchSSHTerminal(chosenLoginService, overrideUsername);
    }

    public void launchSSHTerminal(BasicStringRspec.LoginService loginService) {
        launchSSHTerminal(loginService, null);
    }


    public void launchSSHTerminal(BasicStringRspec.LoginService loginService, String overrideUsername) {
        Preconditions.checkNotNull(loginService);

        if (!externalSshTerminal.isConfigured()) {
            LOG.warn("ExternalSsshTerminal is not configured. Showing warning dialog.");
            JFDialogs.create()
                    .message("jFed must be configured with the location of the terminal application and a valid Private Key.")
                    .showWarning();
            preferencesDialogFactory.showDialog();
        }
        if (!externalSshTerminal.isConfigured()) {
            JFDialogs.create()
                    .message("jFed was not properly configured. Please configure jFed first, and try again.")
                    .showError();
            return;
        }

        externalSshTerminal.launch(
                overrideUsername != null ? overrideUsername : loginService.getUsername(),
                loginService.getHostname(),
                loginService.getPort(),
                null);
    }

    private void launchSSHTerminal(@Nonnull LoginService loginService,
                                   @Nullable JFedConnection.SshProxyInfo proxyInfo,
                                   @Nullable String overrideUsername) {

        if (!externalSshTerminal.isConfigured()) {
            LOG.warn("ExternalSsshTerminal is not configured. Showing warning dialog.");
            JFDialogs.create()
                    .message("jFed must be configured with the location of the terminal application and a valid Private Key.")
                    .showWarning();
            preferencesDialogFactory.showDialog();
        }
        if (!externalSshTerminal.isConfigured()) {
            JFDialogs.create()
                    .message("jFed was not properly configured. Please configure jFed first, and try again.")
                    .showError();
            return;
        }

        externalSshTerminal.launch(
                overrideUsername != null ? overrideUsername : loginService.getUsername(),
                loginService.getHostname(),
                loginService.getIntPort(),
                proxyInfo);

    }


}
