package be.iminds.ilabt.jfed.lowlevel.api;

import be.iminds.ilabt.jfed.log.Logger;
import be.iminds.ilabt.jfed.lowlevel.*;
import be.iminds.ilabt.jfed.lowlevel.connection.SfaConnection;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import org.slf4j.LoggerFactory;

import java.util.Hashtable;
import java.util.Vector;

/**
 * AbstractGeniAggregateManager
 */
public abstract class AbstractGeniAggregateManager extends AbstractApi {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(AbstractGeniAggregateManager.class);

    public AbstractGeniAggregateManager(Logger logger, RetrySettings retrySettings, ServerType serverType, JFedPreferences jFedPreferences) {
        super(logger, retrySettings, serverType, jFedPreferences);
    }

    @Override
    protected boolean isBusyReply(be.iminds.ilabt.jfed.lowlevel.XMLRPCCallDetails res) {
        if (res instanceof XMLRPCCallDetailsWithCodeValueError) {
            XMLRPCCallDetailsWithCodeValueError geniDetails = (XMLRPCCallDetailsWithCodeValueError) res;

            if (geniDetails.getResultCode() == null ||
                    !(geniDetails.getResultCode() instanceof Hashtable))
                return false;

            Hashtable codeStruct = (Hashtable) geniDetails.getResultCode();
            int code = (Integer) codeStruct.get("geni_code");

            if (GeniAMResponseCode.getByCode(code).isBusy())
                return true;

            //exception:  (see FEDIBBTDEV-980)
            //   ION might return geni_code 5 (server error) when renew is not possible, with output text:
            //       ": RenewSliver: Internal API error: RSpec 'urn:publicid:IDN+ch.geni.net:CHtest+slice+mix1' is in 'VLANS-CREATED' status -- cannot renew for now!"
            //   We will treat this as "busy"
            if (code == 5 &&
                    geniDetails.getResultOutput() != null &&
                    geniDetails.getResultOutput().contains("cannot renew for now"))
                return true;
        }
        return false;
    }

    @Override
    protected boolean isPossibleTemporaryErrorReply(be.iminds.ilabt.jfed.lowlevel.XMLRPCCallDetails res) {
        if (res instanceof XMLRPCCallDetailsWithCodeValueError) {
            XMLRPCCallDetailsWithCodeValueError geniDetails = (XMLRPCCallDetailsWithCodeValueError) res;
            if (!(geniDetails.getResultCode() instanceof Hashtable))
                return false;

            Hashtable codeStruct = (Hashtable) geniDetails.getResultCode();

            if (codeStruct == null || !codeStruct.containsKey("geni_code"))
                return false;

            int code = (Integer) codeStruct.get("geni_code");

            if (code == 3 &&
                    geniDetails.getResultOutput().trim().equals("Credential owner does not match speaksfor target")) {
                LOG.warn("Received \"Credential owner does not match speaksfor target\" error. Will handle this as \"possible temporary error\" and try again.");
                return true;
            }
        }
        return false;
    }

    public abstract AggregateManagerReply<? extends AbstractVersionInfo> getVersion(SfaConnection con) throws JFedException;

    public static class AggregateManagerReply<T> implements SfaApiCallReply<T> {
        private GeniAMResponseCode genicode;
        private T val;
        /* output is typically set only on error*/
        private String output;

        //optional
        private boolean hasAmTypeBool;
        private boolean hasAmCodeBool;
        private String amType;
        private int amCode;

        private Hashtable rawResult;
        private XMLRPCCallDetailsWithCodeValueError xMLRPCCallDetailsWithCodeValueError;

        protected AggregateManagerReply() {
            //init self
        }

        public AggregateManagerReply(XMLRPCCallDetailsWithCodeValueError res) {
            this.xMLRPCCallDetailsWithCodeValueError = res;
            this.rawResult = res.getResult();

            Hashtable r = res.getResult();
            try {
                this.val = (T) r.get("value");
            } catch (/*ClassCast*/Exception e) {
                //ignore
                this.val = null;
            }
            Hashtable codeStruct = (Hashtable) res.getResultCode();
            int code = codeStruct == null ? GeniAMResponseCode.SERVER_REPLY_ERROR.getCode()
                    : (Integer) codeStruct.get("geni_code");
            this.genicode = GeniAMResponseCode.getByCode(code);
            this.output = null; //this should be a string, but we allow more. This may be null on success.
            if (r != null && r.get("output") != null) {
                this.output = r.get("output").toString();
                //any empty structur eis interpreted as an empty string
                if (r.get("output") instanceof Hashtable && ((Hashtable) r.get("output")).isEmpty())
                    this.output = "";
                if (r.get("output") instanceof Vector && ((Vector) r.get("output")).isEmpty())
                    this.output = "";
            }
            this.hasAmTypeBool = codeStruct == null ? false : codeStruct.containsKey("am_type");
            this.hasAmCodeBool = codeStruct == null ? false : codeStruct.containsKey("am_code");
            if (hasAmTypeBool)
                if (codeStruct.get("am_type") instanceof String)
                    this.amType = (String) codeStruct.get("am_type");
                else {
                    LOG.warn("am_type in geni_code is not a String but a " + codeStruct.get("am_type").getClass().getName());
                }
            if (hasAmCodeBool) {
                if (codeStruct.get("am_code") instanceof Integer)
                    this.amCode = (Integer) codeStruct.get("am_code");
                else {
                    LOG.warn("am_code in geni_code is not an Integer but a " + codeStruct.get("am_code").getClass().getName());
                }
            }
        }

        public AggregateManagerReply(XMLRPCCallDetailsWithCodeValueError res, T val) {
            this(res);
            this.val = val;
        }

        public static boolean isSuccess(be.iminds.ilabt.jfed.lowlevel.XMLRPCCallDetails res) {
            if (res instanceof XMLRPCCallDetailsWithCodeValueError) {
                XMLRPCCallDetailsWithCodeValueError geniDetails = (XMLRPCCallDetailsWithCodeValueError) res;
                Hashtable codeStruct = (Hashtable) geniDetails.getResultCode();
                int code = (Integer) codeStruct.get("geni_code");
                GeniResponseCode genicode = GeniAMResponseCode.getByCode(code);
                return genicode.isSuccess();
            }
            return false; //not a geni reply, so not successful
        }

        public int getCode() {
            return genicode.getCode();
        }

        public GeniAMResponseCode getGeniResponseCode() {
            return genicode;
        }

        public T getValue() {
            return val;
        }

        /**
         * can be null
         */
        public String getOutput() {
            return output;
        }

        @Override
        public XMLRPCCallDetailsWithCodeValueError getXMLRPCCallDetailsWithCodeValueError() { return xMLRPCCallDetailsWithCodeValueError; }
        @Override
        public XMLRPCCallDetails getXMLRPCCallDetails() {
            return xMLRPCCallDetailsWithCodeValueError;
        }

        public Hashtable getRawResult() {
            return rawResult;
        }

        public Object getRawValue() {
            if (rawResult == null)
                return null;
            return rawResult.get("value");
        }

        public boolean hasAmType() {
            return hasAmTypeBool;
        }

        public boolean hasAmCode() {
            return hasAmCodeBool;
        }

        public String getAmType() {
            if (!hasAmTypeBool)
                throw new RuntimeException("AggregateManagerReply does not have am_type");
            return amType;
        }

        public void setAmType(String amType) {
            hasAmTypeBool = true;
            this.amType = amType;
        }

        public int getAmCode() {
            if (!hasAmCodeBool)
                throw new RuntimeException("AggregateManagerReply does not have am_code");
            return amCode;
        }

        public void setAmCode(int amCode) {
            hasAmCodeBool = true;
            this.amCode = amCode;
        }
    }

    public static abstract class AbstractVersionInfo {
        public abstract int getApi();
    }

    public static boolean isNotEnoughFreeResources(AggregateManagerReply reply) {
        if (reply == null)
            return false;
        if (reply.getGeniResponseCode().isSuccess())
            return false;
        if (reply.getGeniResponseCode().isBusy())
            return false;
        if (reply.getOutput() == null)
            return false;

        String output = reply.getOutput().toLowerCase();

        //Some AM's, like emulab, add a string value when an error occurs. It can't hurt to include that in this search as well.
        if (reply.getValue() != null && reply.getValue() instanceof String)
            output += ((String) reply.getValue()).toLowerCase();
        else {
            if (reply.getRawValue() != null && reply.getRawValue() instanceof String)
                output += ((String) reply.getRawValue()).toLowerCase();
        }

        if (output.contains("no available resources".toLowerCase())) //emulab and instageni
            return true;
        if (output.contains("could not map to resources".toLowerCase()))  //exogeni     (also emulab in value instead of output)
            return true;
        if (output.contains("free resources".toLowerCase()))
            return true;
        if (output.contains("failed to map".toLowerCase()))
            return true;
        if (output.contains("available resources".toLowerCase()))
            return true;
        if (output.contains("no possible mapping".toLowerCase())) //emulab   (example: " *** No possible mapping for PC\n")
            return true;
        if (output.contains("ERROR: mapper: Unretriable error. Giving up".toLowerCase()))
            return true;
        if (output.contains("No available physical nodes of type ".toLowerCase()))
            return true;
        if (output.contains("insufficient resources to fulfill request".toLowerCase()))
            return true;
        if (output.contains("Not enough resources".toLowerCase())) //exogeni:  FEDIBBTDEV-2033
            return true;
        if (output.contains("insufficient resources available".toLowerCase())) //log-a-tec
            return true;

        //bonfire:
        if (output.contains("Got http code 400 when trying to allocate resources".toLowerCase())
                && output.contains("Couldn't create a reservation".toLowerCase()))
            return true;

        return false;
    }
}
