package be.iminds.ilabt.jfed.lowlevel.api;

import be.iminds.ilabt.jfed.log.Logger;
import be.iminds.ilabt.jfed.lowlevel.*;
import be.iminds.ilabt.jfed.lowlevel.connection.HttpConnection;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import org.apache.http.entity.ContentType;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

/**
 * OCCI A limited OCCI client implementation, for use with Bonfire
 * <p/>
 * (can always be extended later for full OCCI support)
 */
public class OCCI extends RestApi {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(OCCI.class);

    /**
     * A human readable name for the implemented API
     *
     * @return "OCCI (limited)";
     */
    static public String getApiName() {
        return "OCCI (limited)";
    }

    /**
     * A human readable name for the implemented API
     *
     * @return "OCCI (limited)";
     */
    @Override
    public String getName() {
        return getApiName();
    }

    /**
     * @param logger the logger to use. May be null if no logger is used.
     */
    public OCCI(@Nonnull Logger logger, @Nonnull JFedPreferences jFedPreferences) {
        super(logger, new ServerType(ServerType.GeniServerRole.OCCI, 1), jFedPreferences);
    }

    @Override
    protected boolean isBusyReply(XMLRPCCallDetails res) {
        return false;
    }

    @ApiMethod(order = 1, hint = "GET /experiments")
    public RestApi.RestReply<String> getExperiments(HttpConnection con) throws JFedException {
        String javaName = "getExperiments";
        String relUrl = "/experiments";
        Map<String, Object> methodParams = makeMethodParameters();

        return internalGet(con, javaName, relUrl, methodParams, methodParams);
    }

    @ApiMethod(order = 2, hint = "GET /experiments/<ID>")
    public RestApi.RestReply<String> getExperiment(HttpConnection con,
                                           @ApiMethodParameter(name = "experimentId", parameterType = ApiMethodParameterType.STRING, hint = "")
                                           String experimentId) throws JFedException {
        String javaName = "getExperiment";
        String relUrl = "/experiments/" + experimentId;
        Map<String, Object> methodParams = makeMethodParameters("experimentId", experimentId);

        return internalGet(con, javaName, relUrl, methodParams, methodParams);
    }

    @ApiMethod(order = 3, hint = "POST /experiments")
    public RestApi.RestReply<String> postExperiment(HttpConnection con,
                                            @ApiMethodParameter(name = "xml", parameterType = ApiMethodParameterType.STRING_MULTILINE, hint = "",
                                                    guiDefault = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n" +
                                                            "<experiment xmlns=\"http://api.bonfire-project.eu/doc/schemas/occi\">\n" +
                                                            "<name>NAME</name>\n" +
                                                            "<groups>GROUPS</groups>\n" +
                                                            "<description>DESCRIPTION</description>\n" +
                                                            "<walltime>WALLTIME</walltime>\n" +
                                                            "</experiment>")
                                            String xml) throws JFedException {
        String javaName = "postExperiment";
        String relUrl = "/experiments";
        Map<String, Object> methodParams = makeMethodParameters("xml", xml);

        return internalPost(con, javaName, relUrl, methodParams, methodParams, ContentType.create("application/vnd.bonfire+xml", "UTF-8"), xml);
    }

    @ApiMethod(order = 4, hint = "POST /experiments")
    public RestApi.RestReply<String> postExperimentWizard(HttpConnection con,
                                                  @ApiMethodParameter(name = "groups", parameterType = ApiMethodParameterType.STRING, hint = "", required = false)
                                                  String groups,
                                                  @ApiMethodParameter(name = "name", parameterType = ApiMethodParameterType.STRING, hint = "")
                                                  String name,
                                                  @ApiMethodParameter(name = "wallTime", parameterType = ApiMethodParameterType.INTEGER, hint = "")
                                                  int wallTime,
                                                  @ApiMethodParameter(name = "description", parameterType = ApiMethodParameterType.STRING, hint = "", required = false)
                                                  String description) throws JFedException {
        String javaName = "postExperiment";
        String relUrl = "/experiments";
        Map<String, Object> methodParams =
                makeMethodParameters("groups", groups, "name", name, "wallTime", wallTime, "description", description);

        String experimentXmlString = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n" +
                "<experiment xmlns=\"http://api.bonfire-project.eu/doc/schemas/occi\">\n" +
                (groups == null ? "" : "<groups>" + groups + "</groups>") +
                "  <name>" + name + "</name>\n" +
                "  <walltime>" + wallTime + "</walltime>\n" +
                (description == null ? "" : "  <description>" + description + "</description>\n") +
                "</experiment>";

        return internalPost(con, javaName, relUrl, methodParams, methodParams, ContentType.create("application/vnd.bonfire+xml", "UTF-8"), experimentXmlString);
    }

//    @ApiMethod(order=5, hint="DELETE /experiments/<ID>")
//    public OCCIReply<String> updateExperimentWallTime(HttpConnection con,
//                                                      @ApiMethodParameter(name="experimentId", parameterType=ApiMethodParameterType.STRING, hint="")
//                                                      String experimentId,
//                                                      @ApiMethodParameter(name="wallTime", parameterType=ApiMethodParameterType.INTEGER, hint="")
//                                                      int wallTime) throws JFedException {
//        String javaName = "updateExperimentWallTime";
//        String relUrl = "/experiments/"+experimentId;
//        Map<String, Object> methodParams = makeMethodParameters("experimentId", experimentId, "wallTime", wallTime);
//
//        String experimentXmlString = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n" +
//                "<experiment xmlns=\"http://api.bonfire-project.eu/doc/schemas/occi\">\n" +
//                "  <walltime>"+wallTime+"</walltime>\n" +
//                "</experiment>";
//
    //TODO: check how to update wall time

//        return internalPut(con, javaName, relUrl, methodParams, ContentType.create("application/vnd.bonfire+xml", "UTF-8"), experimentXmlString);
//    }


    @ApiMethod(order = 5, hint = "DELETE /experiments/<ID>")
    public RestApi.RestReply<String> deleteExperiment(HttpConnection con,
                                              @ApiMethodParameter(name = "experimentId", parameterType = ApiMethodParameterType.STRING, hint = "")
                                              String experimentId) throws JFedException {
        String javaName = "deleteExperiment";
        String relUrl = "/experiments/" + experimentId;
        Map<String, Object> methodParams = makeMethodParameters("experimentId", experimentId);

        return internalDelete(con, javaName, relUrl, methodParams, methodParams);
    }


    @ApiMethod(order = 10, hint = "GET /experiments/<ID>/computes")
    public RestApi.RestReply<String> getExperimentComputes(HttpConnection con,
                                                   @ApiMethodParameter(name = "experimentId", parameterType = ApiMethodParameterType.STRING, hint = "")
                                                   String experimentId) throws JFedException {
        String javaName = "getExperimentComputes";
        String relUrl = "/experiments/" + experimentId + "/computes";
        Map<String, Object> methodParams = makeMethodParameters("experimentId", experimentId);

        return internalGet(con, javaName, relUrl, methodParams, methodParams);
    }

    @ApiMethod(order = 12, hint = "POST /experiments/<ID>/computes")
    public RestApi.RestReply<String> postExperimentCompute(HttpConnection con,
                                                   @ApiMethodParameter(name = "experimentId", parameterType = ApiMethodParameterType.STRING, hint = "")
                                                   String experimentId,
                                                   @ApiMethodParameter(name = "xml", parameterType = ApiMethodParameterType.STRING_MULTILINE,
                                                           guiDefault = "<compute xmlns=\"http://api.bonfire-project.eu/doc/schemas/occi\">\n" +
                                                                   "<name>NAME</name>\n" +
                                                                   "<groups>GROUPS</groups>\n" +
                                                                   "<instance_type>lite</instance_type>\n" +
                                                                   "<disk>\n" +
                                                                   "<storage href=\"/locations/LOCATION/storages/STORAGEID\"/>\n" +
                                                                   "<type>OS</type>\n" +
                                                                   "<target>hda</target>\n" +
                                                                   "</disk>\n" +
                                                                   "<nic>\n" +
                                                                   "<network href=\"/locations/LOCATION/networks/NETWORKID\"/>\n" +
                                                                   "</nic>\n" +
                                                                   "<link href=\"/locations/LOCATION\" rel=\"location\"/>\n" +
                                                                   "</compute>", hint = "")
                                                   String xml) throws JFedException {
        String javaName = "postExperimentCompute";
        String relUrl = "/experiments/" + experimentId + "/computes";
        Map<String, Object> methodParams = makeMethodParameters("experimentId", experimentId, "xml", xml);

        return internalPost(con, javaName, relUrl, methodParams, methodParams, ContentType.create("application/vnd.bonfire+xml", "UTF-8"), xml);
    }

    @ApiMethod(order = 13, hint = "POST /experiments/<ID>/computes")
    public RestApi.RestReply<String> postExperimentComputeWizard(HttpConnection con,
                                                         @ApiMethodParameter(name = "experimentId", parameterType = ApiMethodParameterType.STRING, hint = "")
                                                         String experimentId,
                                                         @ApiMethodParameter(name = "computeName", parameterType = ApiMethodParameterType.STRING, hint = "")
                                                         String computeName,
                                                         @ApiMethodParameter(name = "groups", required = false, parameterType = ApiMethodParameterType.STRING, hint = "")
                                                         String groups,
                                                         @ApiMethodParameter(name = "locationName", parameterType = ApiMethodParameterType.STRING, hint = "")
                                                         String locationName,
                                                         @ApiMethodParameter(name = "instanceType", parameterType = ApiMethodParameterType.STRING, guiDefault = "lite", hint = "")
                                                         String instanceType,
                                                         @ApiMethodParameter(name = "networkId", parameterType = ApiMethodParameterType.STRING, hint = "")
                                                         String networkId,
                                                         @ApiMethodParameter(name = "storageId", parameterType = ApiMethodParameterType.STRING, hint = "")
                                                         String storageId) throws JFedException {
        List<String> networkIds = new ArrayList<String>();
        if (networkId != null) networkIds.add(networkId);

        List<String> storageIds = new ArrayList<String>();
        if (storageId != null) storageIds.add(storageId);

        return postExperimentComputeWizard(con, experimentId, computeName, groups, locationName, instanceType, networkIds, storageIds);
    }

    @ApiMethod(order = 14, hint = "POST /experiments/<ID>/computes")
    public RestApi.RestReply<String> postExperimentComputeWizard(HttpConnection con,
                                                         @ApiMethodParameter(name = "experimentId", parameterType = ApiMethodParameterType.STRING, hint = "")
                                                         String experimentId,
                                                         @ApiMethodParameter(name = "computeName", parameterType = ApiMethodParameterType.STRING, hint = "")
                                                         String computeName,
                                                         @ApiMethodParameter(name = "groups", required = false, parameterType = ApiMethodParameterType.STRING, hint = "")
                                                         String groups,
                                                         @ApiMethodParameter(name = "locationName", parameterType = ApiMethodParameterType.STRING, hint = "")
                                                         String locationName,
                                                         @ApiMethodParameter(name = "instanceType", parameterType = ApiMethodParameterType.STRING, guiDefault = "lite", hint = "")
                                                         String instanceType,
                                                         @ApiMethodParameter(name = "networkIds", parameterType = ApiMethodParameterType.LIST_OF_STRING, hint = "")
                                                         List<String> networkIds,
                                                         @ApiMethodParameter(name = "storageIds", parameterType = ApiMethodParameterType.LIST_OF_STRING, hint = "")
                                                         List<String> storageIds) throws JFedException {
        String javaName = "postExperimentComputeWizard";
        String relUrl = "/experiments/" + experimentId + "/computes";
        Map<String, Object> methodParams = makeMethodParameters("experimentId", experimentId,
                "computeName", computeName,
                "groups", groups,
                "locationName", locationName,
                "instanceType", instanceType,
                "networkIds", networkIds,
                "storageIds", storageIds);

        String networkIdsString = "";
        String storageIdsString = "";
        int nicId = 0;
        for (String networkId : networkIds) {
            networkIdsString += "  <nic id=\"" + (nicId++) + "\">\n" +
                    "    <network href=\"https://api.bonfire-project.eu/locations/" + locationName + "/networks/" + networkId + "\"/>\n" +
                    "  </nic>\n";
        }
        int diskId = 0;
        for (String storageId : storageIds) {
            storageIdsString += "  <disk id=\"" + (diskId++) + "\">\n" +
                    "    <type>OS</type>\n" +
                    "    <storage href=\"https://api.bonfire-project.eu/locations/" + locationName + "/storages/" + storageId + "\"/>\n" +
                    "  </disk>\n";
        }

        String xmlString = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n" +
                "<compute xmlns=\"http://api.bonfire-project.eu/doc/schemas/occi\">\n" +
                "  <name>" + computeName + "</name>\n" +
                (groups == null ? "" : "  <groups>" + groups + "</groups>\n") +
                "  <location href=\"https://api.bonfire-project.eu/locations/" + locationName + "\"/>\n" +
                "  <instance_type>" + instanceType + "</instance_type>\n" +
                networkIdsString +
                storageIdsString +
                "</compute>\n";

        return internalPost(con, javaName, relUrl, methodParams, methodParams, ContentType.create("application/vnd.bonfire+xml", "UTF-8"), xmlString);
    }

    @ApiMethod(order = 20, hint = "GET /locations")
    public RestApi.RestReply<String> getLocations(HttpConnection con) throws JFedException {
        String javaName = "getLocations";
        String relUrl = "/locations";
        Map<String, Object> methodParams = makeMethodParameters();

        return internalGet(con, javaName, relUrl, methodParams, methodParams);
    }

    @ApiMethod(order = 21, hint = "GET /locations/<ID>")
    public RestApi.RestReply<String> getLocation(HttpConnection con,
                                         @ApiMethodParameter(name = "locationId", parameterType = ApiMethodParameterType.STRING, hint = "")
                                         String locationName) throws JFedException {
        String javaName = "getLocation";
        String relUrl = "/locations/" + locationName;
        Map<String, Object> methodParams = makeMethodParameters("locationName", locationName);

        return internalGet(con, javaName, relUrl, methodParams, methodParams);
    }

    @ApiMethod(order = 30, hint = "GET /locations/<ID>/storages")
    public RestApi.RestReply<String> getLocationStorages(HttpConnection con,
                                                 @ApiMethodParameter(name = "locationName", parameterType = ApiMethodParameterType.STRING, hint = "")
                                                 String locationName) throws JFedException {
        String javaName = "getLocationStorages";
        String relUrl = "/locations/" + locationName + "/storages";
        Map<String, Object> methodParams = makeMethodParameters("locationName", locationName);

        return internalGet(con, javaName, relUrl, methodParams, methodParams);
    }

    @ApiMethod(order = 31, hint = "GET /locations/<ID>/storages/<ID>")
    public RestApi.RestReply<String> getLocationStorage(HttpConnection con,
                                                @ApiMethodParameter(name = "locationName", parameterType = ApiMethodParameterType.STRING, hint = "")
                                                String locationName,
                                                @ApiMethodParameter(name = "storageId", parameterType = ApiMethodParameterType.STRING, hint = "")
                                                String storageId) throws JFedException {
        String javaName = "getLocationStorage";
        String relUrl = "/locations/" + locationName + "/storages/" + storageId;
        Map<String, Object> methodParams = makeMethodParameters("locationName", locationName, "storageId", storageId);

        return internalGet(con, javaName, relUrl, methodParams, methodParams);
    }

    @ApiMethod(order = 40, hint = "GET /locations/<ID>/networks")
    public RestApi.RestReply<String> getLocationNetworks(HttpConnection con,
                                                 @ApiMethodParameter(name = "locationName", parameterType = ApiMethodParameterType.STRING, hint = "")
                                                 String locationName) throws JFedException {
        String javaName = "getLocationNetworks";
        String relUrl = "/locations/" + locationName + "/networks";
        Map<String, Object> methodParams = makeMethodParameters("locationName", locationName);

        return internalGet(con, javaName, relUrl, methodParams, methodParams);
    }

    @ApiMethod(order = 41, hint = "GET /locations/<ID>/networks/<ID>")
    public RestApi.RestReply<String> getLocationNetwork(HttpConnection con,
                                                @ApiMethodParameter(name = "locationName", parameterType = ApiMethodParameterType.STRING, hint = "")
                                                String locationName,
                                                @ApiMethodParameter(name = "networkId", parameterType = ApiMethodParameterType.STRING, hint = "")
                                                String networkId) throws JFedException {
        String javaName = "getLocationNetworks";
        String relUrl = "/locations/" + locationName + "/networks/" + networkId;
        Map<String, Object> methodParams = makeMethodParameters("locationName", locationName, "networkId", networkId);

        return internalGet(con, javaName, relUrl, methodParams, methodParams);
    }

    @ApiMethod(order = 50, hint = "GET /locations/<ID>/computes/<ID>")
    public RestApi.RestReply<String> getLocationCompute(HttpConnection con,
                                                @ApiMethodParameter(name = "locationName", parameterType = ApiMethodParameterType.STRING, hint = "")
                                                String locationName,
                                                @ApiMethodParameter(name = "computeId", parameterType = ApiMethodParameterType.STRING, hint = "")
                                                String computeId) throws JFedException {
        String javaName = "getLocationCompute";
        String relUrl = "/locations/" + locationName + "/computes/" + computeId;
        Map<String, Object> methodParams = makeMethodParameters("locationName", locationName, "computeId", computeId);

        return internalGet(con, javaName, relUrl, methodParams, methodParams);
    }

    @ApiMethod(order = 51, hint = "DELETE /locations/<ID>/computes/<ID>")
    public RestApi.RestReply<String> deleteLocationCompute(HttpConnection con,
                                                   @ApiMethodParameter(name = "locationName", parameterType = ApiMethodParameterType.STRING, hint = "")
                                                   String locationName,
                                                   @ApiMethodParameter(name = "computeId", parameterType = ApiMethodParameterType.STRING, hint = "")
                                                   String computeId) throws JFedException {
        String javaName = "deleteLocationCompute";
        String relUrl = "/locations/" + locationName + "/computes/" + computeId;
        Map<String, Object> methodParams = makeMethodParameters("locationName", locationName, "computeId", computeId);

        return internalDelete(con, javaName, relUrl, methodParams, methodParams);
    }
}
