package be.iminds.ilabt.jfed.lowlevel.connection;

import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.util.*;
import org.apache.http.HttpHost;
import org.apache.http.auth.AuthScope;
import org.apache.http.auth.UsernamePasswordCredentials;
import org.apache.http.client.AuthCache;
import org.apache.http.client.protocol.ClientContext;
import org.apache.http.conn.ClientConnectionManager;
import org.apache.http.conn.scheme.Scheme;
import org.apache.http.conn.scheme.SchemeRegistry;
import org.apache.http.conn.scheme.SchemeSocketFactory;
import org.apache.http.conn.ssl.BrowserCompatHostnameVerifier;
import org.apache.http.conn.ssl.SSLSocketFactory;
import org.apache.http.conn.ssl.TrustStrategy;
import org.apache.http.impl.auth.BasicScheme;
import org.apache.http.impl.client.BasicAuthCache;
import org.apache.http.impl.client.DefaultHttpClient;
import org.apache.http.impl.conn.SingleClientConnManager;
import org.apache.http.params.BasicHttpParams;
import org.apache.http.params.CoreConnectionPNames;
import org.apache.http.params.HttpParams;
import org.apache.http.protocol.BasicHttpContext;
import org.apache.http.protocol.HttpContext;
import org.apache.xmlrpc.XmlRpcClient;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.net.ssl.SSLContext;
import java.net.InetSocketAddress;
import java.net.MalformedURLException;
import java.net.URL;
import java.security.*;
import java.security.cert.*;
import java.security.cert.Certificate;
import java.util.*;

/**
 * ConnectionFactory
 */
public class BasicConnectionBuilder implements ConnectionBuilder {
    private static final Logger LOG = LoggerFactory.getLogger(BasicConnectionBuilder.class);

    private URL serverUrl;
    private String serverUrlString;
    private JFedConnection.DebugInfo debugInfo;
    private JFedConnection.ProxyInfo proxyInfo;
    private ConnectionConfig.Authentication authentication;
    private JFedPasswordManager.LoginInfo login;
    private List<X509Certificate> clientCertChain;
    private PrivateKey privateKey;
    private ConnectionConfig.Protocol protocol;
    private JFedTrustStore conTrustStore;
    private HandleUntrustedCallback handleUntrustedCallback;

    private boolean hackAllowAllServerCertificates = false;

    /** class private. In fact, should only be called from BasicConnectionBuilderFactory */
    BasicConnectionBuilder() {

    }

    @Override
    public void setUrl(String serverUrlString) throws MalformedURLException {
        this.serverUrlString = serverUrlString;
        this.serverUrl = new URL(serverUrlString);
    }

    @Override
    public void setUrl(URL serverUrl) {
        this.serverUrl = serverUrl;
        this.serverUrlString = serverUrl.toExternalForm();
    }

    @Override
    public void setHackAllowAllServerCertificates(boolean hackAllowAllServerCertificates) {
        this.hackAllowAllServerCertificates = hackAllowAllServerCertificates;
    }

    @Override
    public void setDebugInfo(JFedConnection.DebugInfo debugInfo) {
        this.debugInfo = debugInfo;
    }

    @Override
    public void setProxy(JFedConnection.ProxyInfo proxyInfo) {
        this.proxyInfo = proxyInfo;
    }

    @Override
    public void useNoAuthentication() {
        this.authentication = ConnectionConfig.Authentication.NONE;
    }

    @Override
    public void useHttpBasicAuthentication(JFedPasswordManager.LoginInfo login) {
        this.authentication = ConnectionConfig.Authentication.HTTP_BASIC;
        this.login = login;
        if (login == null)
            throw new IllegalArgumentException("Illegal argument: login == null");
    }

    @Override
    public void useSslClientAuthentication(List<X509Certificate> clientCertChain, PrivateKey privateKey) {
        this.authentication = ConnectionConfig.Authentication.SSL_CLIENT_AUTH;
        this.clientCertChain = clientCertChain;
        this.privateKey = privateKey;
        if (privateKey == null)
            throw new IllegalArgumentException("Illegal argument: privateKey == null");
        if (clientCertChain == null)
            throw new IllegalArgumentException("Illegal argument: clientCertificateChain == null");
        if (clientCertChain.isEmpty())
            throw new IllegalArgumentException("Illegal argument: clientCertChain is empty");
    }

    @Override
    public void useHttp() {
        this.protocol = ConnectionConfig.Protocol.HTTP;
    }

    @Override
    public void useHttps(JFedTrustStore conTrustStore, HandleUntrustedCallback handleUntrustedCallback) {
        this.protocol = ConnectionConfig.Protocol.HTTPS;
        this.conTrustStore = conTrustStore;
        this.handleUntrustedCallback = handleUntrustedCallback;
        if (conTrustStore == null)
            throw new IllegalArgumentException("Illegal argument: conTrustStore == null");
    }

    private void check() {
        if (debugInfo == null) throw new RuntimeException("Configuration problem: must specify debugInfo");
        if (serverUrlString == null || serverUrl == null) throw new RuntimeException("Configuration problem: must specify URL");
        if (authentication == null) throw new RuntimeException("Configuration problem: must specify authentication");
        if (protocol == null) throw new RuntimeException("Configuration problem: must specify HTTP or HTTPS");
        if (protocol != ConnectionConfig.Protocol.HTTPS && authentication == ConnectionConfig.Authentication.SSL_CLIENT_AUTH)
            throw new RuntimeException("Configuration problem: incompatible options: cannot do SSL client authentication without HTTPS.");
    }


    private class HttpConParts {
        final DefaultHttpClient defaultHttpClient;
        final HttpContext httpContext;
        final HttpContext preemptiveAuthHttpContext;
        final TrustStrategy trustStrategy;

        public HttpConParts(DefaultHttpClient defaultHttpClient, HttpContext httpContext, HttpContext preemptiveAuthHttpContext, TrustStrategy trustStrategy) {
            this.defaultHttpClient = defaultHttpClient;
            this.httpContext = httpContext;
            this.preemptiveAuthHttpContext = preemptiveAuthHttpContext;
            this.trustStrategy = trustStrategy;
        }
    }

    private static class SpyingTrustStrategy implements TrustStrategy {
        private final TrustStrategy orig;
        public X509Certificate[] chain;
        private JFedConnection connection;

        public SpyingTrustStrategy(TrustStrategy orig) {
            this.orig = orig;
        }

        @Override
        public boolean isTrusted(X509Certificate[] certificate, String authType) throws CertificateException {
            if (certificate != null && certificate.length > 0) {
                chain = certificate;
                if (connection != null)
                    connection.registerServerCertficates(chain);
            }
            return orig.isTrusted(certificate, authType);
        }

        public void setConnection(JFedConnection connection) {
            LOG.debug("registering connection to SpyingTrustStrategy");
            this.connection = connection;
            if (chain != null)
                connection.registerServerCertficates(chain);
        }
    }

    private HttpConParts getConnectionHttpClientAndContext() throws JFedException {
        HttpParams params = new BasicHttpParams();
        params.setParameter(CoreConnectionPNames.SO_TIMEOUT, 120000); //read timeout in ms (default: 0 = infinite)   (long timeout, as some calls take time to finish)
        params.setParameter(CoreConnectionPNames.CONNECTION_TIMEOUT, 10000); //connection timeout in ms (default: 0 = infinite)

        DefaultHttpClient httpClient;

        if (System.getProperty("proxySet") != null &&
                System.getProperty("proxySet").equals("true") &&
                System.getProperty("socksProxyHost") != null) {
            if (hackAllowAllServerCertificates) {
                LOG.error("hackAllowAllServerCertificates is not supported with socks proxy. Will not use it!");
            }
            httpClient = SocksProxyHelper.getHttpClientOverSocksProxy(params);
        } else {
            if (hackAllowAllServerCertificates) {
                try {
                    SSLContext sslcontext = SSLContext.getInstance("TLS");
                    sslcontext.init(null, null, null);
                    SSLSocketFactory sf = new SSLSocketFactory(
                            sslcontext,
                            SSLSocketFactory.ALLOW_ALL_HOSTNAME_VERIFIER);
                    Scheme httpsScheme = new Scheme("https", 443, sf);
                    SchemeRegistry schemeRegistry = new SchemeRegistry();
                    schemeRegistry.register(httpsScheme);
                    ClientConnectionManager cm = new SingleClientConnManager(schemeRegistry);

                    httpClient = new DefaultHttpClient(cm);
                    httpClient.setDefaultHttpParams(params);
                } catch (NoSuchAlgorithmException e) {
                    LOG.error("Failed ignoring server certs. Will fall back to not using hackAllowAllServerCertificates.", e);
                    httpClient = new DefaultHttpClient(params);
                } catch (KeyManagementException e) {
                    LOG.error("Failed ignoring server certs. Will fall back to not using hackAllowAllServerCertificates.", e);
                    httpClient = new DefaultHttpClient(params);
                }
            } else {
                httpClient = new DefaultHttpClient(params);
            }
        }

        TrustStrategy trustStrategy = null;
        if (protocol == ConnectionConfig.Protocol.HTTPS) {
            try {
                String key_store_pass = null;
                String privatekey_store_pass =
                        "someotherpass"; //password for internal use. We don't really need one as we'll never save this keystore or share it with other objects
                KeyStore keyStore = null;
                final KeyStore trustStore = conTrustStore.getTrustStore();

                if (authentication == ConnectionConfig.Authentication.SSL_CLIENT_AUTH) {
                    key_store_pass = "somepass"; //password for internal use. We don;t really need one as we'll never save this keystore.
                    keyStore = KeyStore.getInstance("JKS", "SUN");
                    keyStore.load(null, key_store_pass.toCharArray());

                    // Create CertificateChain
                    Certificate[] certs = new Certificate[clientCertChain.size()];
                    for (int i = 0; i < clientCertChain.size(); i++) {
                        X509Certificate c = clientCertChain.get(i);
                        certs[i] = c;
                        try {
                            //FEDIBBTDEV-878: check certificate validity
                            c.checkValidity(new Date());
                        } catch (CertificateExpiredException e) {
                            throw new JFedException("Certificate " + i + " (of " + clientCertChain.size() + ") in the user " +
                                    "certificate chain has expired. NotAfter=" + c.getNotAfter() + " now=" + new Date());
                        } catch (CertificateNotYetValidException e) {
                            throw new JFedException("Certificate " + i + " (of " + clientCertChain.size() + ") in the user " +
                                    "certificate chain is not yet valid. NotBefore=" + c.getNotBefore() + " now=" + new Date());
                        }
                    }

                    // storing private key and certificate to keystore
                    keyStore.setKeyEntry("authority", privateKey, privatekey_store_pass.toCharArray()/*password is used to encrypt the key inside the keystore*/, certs);

                    //add all client certificates to trust store as well as keystore!
                    for (int i = 0; i < clientCertChain.size(); i++)
                        trustStore.setCertificateEntry("clientCertificate" + i, clientCertChain.get(i));
                }

                String algo = SSLSocketFactory.TLS;
                java.security.SecureRandom random = new SecureRandom();

                final List<String> allowedCertificateHostnameAliases =
                        new ArrayList<String>(conTrustStore.getAllowedServerCertificateHostnameAliases());

                SSLSocketFactory socketFactory;
                if (handleUntrustedCallback != null) {
                    //                socketFactory = new SSLSocketFactory(TRUSTALL, new AllowAllHostnameVerifier()); //doesn't work as we need client authentication

                    //didn't work for some reason. Probably it's missing client certificate (need to check apache source code of SSLSocketFactory to correctly write this)
                    //                socketFactory = new MyInsecureSSLAcceptAllSocketFactory(algo, keyStore, privatekey_store_pass/*password is used to encrypt the key inside the keystore*/, trustStore, random, hostnameVerifier);

                    //Solved by connecting 2 times. Not best solution, but it works.
//                    SSLCertificateDownloader.SSLCertificateJFedInfo certInfo =
//                            SSLCertificateDownloader.getCertificateInfo(serverUrl,
//                                    (proxyInfo != null && proxyInfo instanceof JFedConnection.SshProxyInfo) ? (JFedConnection.SshProxyInfo) proxyInfo : null);
//                    if (certInfo != null && certInfo.isSelfSigned() != null && certInfo.isSelfSigned()) {
//                        if (handleUntrustedCallback.trust(certInfo)) {
//                            trustStore.setCertificateEntry("allTrustCert" + extraTrustCount, certInfo.getCert());
//                            if (!certInfo.getSubjectMatchesHostname())
//                                allowedCertificateHostnameAliases.add(certInfo.getSubject());
//                        } else {
//                            LOG.info("User does not trust certificate -> Not adding anything to trust store.");
//                        }
//                    }

                    /** This is probably not a good implementation of the default TrustStrategy! Too bad the real default is nowhere to be found. */
                    trustStrategy = new TrustStrategy() {
                        @Override
                        public boolean isTrusted(X509Certificate[] certificates, String authType) throws CertificateException {
                            if (certificates == null || certificates.length == 0)
                                return false;
                            try {
                                for (X509Certificate cert : certificates) {
                                    String alias = trustStore.getCertificateAlias(cert);
                                    if (alias != null)
                                        return true;
                                }
                            } catch (KeyStoreException e) {
                                throw new CertificateException("problem with keystore while checking certificate", e);
                            }
                            SSLCertificateDownloader.SSLCertificateJFedInfo certInfo =
                                    new SSLCertificateDownloader.SSLCertificateJFedInfo(certificates[0/*certificates.length-1*/], serverUrl, serverUrl.getHost());
                            return handleUntrustedCallback.trust(certInfo);
                        }
                    };

                    LOG.debug("using home-made trustStrategy, this is far from perfect! Might be insecure!");
                }

                org.apache.http.conn.ssl.X509HostnameVerifier hostnameVerifier = makeHostNameVerifier(allowedCertificateHostnameAliases);

                if (trustStrategy != null) {
                    LOG.debug("adding SpyingTrustStrategy");
                    trustStrategy = new SpyingTrustStrategy(trustStrategy);
                } else {
                    //too bad there is no default
//                    trustStrategy = new SpyingTrustStrategy(trustStrategy);
                }

                if (proxyInfo != null && proxyInfo instanceof JFedConnection.SshProxyInfo) {
                    JFedConnection.SshProxyInfo sshProxyInfo = (JFedConnection.SshProxyInfo) proxyInfo;

                    LOG.debug("Using SSH Proxy for connection: " + sshProxyInfo.getHostname());

                    SSLSocketFactory sslSocketFactory = new SshServerProxyHelper.SslOverSshProxySocketFactory(
                            new SshServerProxyHelper.SshProxyInfo(
                                    new InetSocketAddress(sshProxyInfo.getHostname(), sshProxyInfo.getPort()),
                                    sshProxyInfo.getUsername(),
                                    sshProxyInfo.getHostKey(),
                                    new String(KeyUtil.privateKeyToAnyPem(sshProxyInfo.getSshKeyInfo().getPrivateKey()))
                            ),
                            algo,
                            keyStore,
                            privatekey_store_pass /* password is used to encrypt the key inside the keystore */,
                            trustStore,
                            random,
                            trustStrategy,
                            hostnameVerifier
                    );
                    socketFactory = sslSocketFactory;
                } else {
                    SSLSocketFactory sslSocketFactory = new SSLSocketFactory(
                            algo,
                            keyStore,
                            privatekey_store_pass/*password is used to encrypt the key inside the keystore*/,
                            trustStore,
                            random,
                            trustStrategy,
                            hostnameVerifier);
                    socketFactory = sslSocketFactory;
                }


                Scheme sch1 = new Scheme("https", 443, socketFactory); //port is the default port for the given protocol
                httpClient.getConnectionManager().getSchemeRegistry().register(sch1);
            } catch (Exception e) {
                throw new JFedException("Error creating SSL connection to " + serverUrlString, e);
            }
        }

        if (protocol == ConnectionConfig.Protocol.HTTP) {
            if (proxyInfo != null && proxyInfo instanceof JFedConnection.SshProxyInfo) {
                JFedConnection.SshProxyInfo sshProxyInfo = (JFedConnection.SshProxyInfo) proxyInfo;

                LOG.debug("Using SSH Proxy HTTP connection: " + sshProxyInfo.getHostname());

                SchemeSocketFactory socketFactory = new SshServerProxyHelper.PlainHttpOverSshProxySocketFactory(
                        new SshServerProxyHelper.SshProxyInfo(
                                new InetSocketAddress(sshProxyInfo.getHostname(), sshProxyInfo.getPort()),
                                sshProxyInfo.getUsername(),
                                sshProxyInfo.getHostKey(),
                                new String(KeyUtil.privateKeyToAnyPem(sshProxyInfo.getSshKeyInfo().getPrivateKey()))
                        )
                );

                Scheme sch1 = new Scheme("http", 80/*port is the default port for the given protocol*/, socketFactory);
                httpClient.getConnectionManager().getSchemeRegistry().register(sch1);
            }
        }

        BasicHttpContext preemptiveAuthHttpContext = null;
        if (authentication == ConnectionConfig.Authentication.HTTP_BASIC) {
            int port = serverUrl.getPort();
            if (serverUrl.getPort() <= 0) port = serverUrl.getDefaultPort();
            httpClient.getCredentialsProvider().setCredentials(
                    new AuthScope(serverUrl.getHost(), port),
                    new UsernamePasswordCredentials(login.getUsername(), login.getPassword()));

            //credentials are now ready, and will be provided when server ask for it.
            //below we will construct a HttpContext that can be used to enable preemptive authentication for specific calls
            //    (= send the authentication before the server asks for it)

            //TODO: Currently, onlt HTTP / REST conenctions support this, SFA does not.

            AuthCache authCache = new BasicAuthCache();
            BasicScheme basicAuth = new BasicScheme();
            authCache.put(new HttpHost(serverUrl.getHost(), port, protocol == ConnectionConfig.Protocol.HTTP ? "http" : "https"), basicAuth);
            preemptiveAuthHttpContext = new BasicHttpContext();
            preemptiveAuthHttpContext.setAttribute(ClientContext.AUTH_CACHE, authCache);
        }

        return new HttpConParts(httpClient, null, preemptiveAuthHttpContext, trustStrategy);
    }

    private static int extraTrustCount = 7000;
    static org.apache.http.conn.ssl.X509HostnameVerifier makeHostNameVerifier(final List<String> allowedCertificateHostnameAliases) {
        return new org.apache.http.conn.ssl.AbstractVerifier() {
            private BrowserCompatHostnameVerifier base = new BrowserCompatHostnameVerifier();

            /**
             * @param host the hostname of the url to check.
             * @param cns the CN's of the certificate
             * @param subjectAlts the aliases fot the CN's that the certificate allows
             * @throws javax.net.ssl.SSLException
             */
            @Override
            public final void verify(java.lang.String host, String[] cns, String[] subjectAlts) throws javax.net.ssl.SSLException {
                List<String> modifiedSubjectAlts = new ArrayList<String>();
                List<String> modifiedCns = new ArrayList<String>();


                if (subjectAlts != null)
                    Collections.addAll(modifiedSubjectAlts, subjectAlts);
                if (cns != null)
                    Collections.addAll(modifiedCns, cns);

                //force add of host itself, this makes any certificate we have stored match
                for (String allowedAlias : allowedCertificateHostnameAliases)
                    if (modifiedCns.contains(allowedAlias) || modifiedSubjectAlts.contains(allowedAlias) || allowedAlias.equals("*")) {
                        modifiedSubjectAlts.clear();
                        modifiedCns.clear();

                        modifiedSubjectAlts.add(host);
                        modifiedCns.add(host);
                        break;
                    }

                String[] newSubjectAlts = new String[modifiedSubjectAlts.size()];
                for (int i = 0; i < modifiedSubjectAlts.size(); i++)
                    newSubjectAlts[i] = modifiedSubjectAlts.get(i);

                String[] newCns = new String[modifiedCns.size()];
                for (int i = 0; i < modifiedCns.size(); i++)
                    newCns[i] = modifiedCns.get(i);

                try {
                    base.verify(host, newCns, newSubjectAlts);
                } catch (javax.net.ssl.SSLException e) {
                    LOG.warn("HttpsClientWithUserAuthenticationFactory hostnameVerifier host name verification failed:" + e.getMessage() + "\n" +
                            "host=" + host + "\n" +
                            "cns=" + (cns == null ? "null" : Arrays.toString(cns)) + "\n" +
                            "subjectAlts=" + Arrays.toString(subjectAlts) + "\n" +
                            "newCns=" + modifiedCns + "\n" +
                            "newSubjectAlts=" + modifiedSubjectAlts+"\n"+
                            "allowedCertificateHostnameAliases=" + allowedCertificateHostnameAliases
                            , e);
                    throw e;
                }
            }
        };
    }

    @Override
    public SfaConnection buildSfaConnection() throws JFedException {
        check();

        HttpConParts httpConParts = getConnectionHttpClientAndContext();
        DefaultHttpClient httpClient = httpConParts.defaultHttpClient;

        CommonsHttpClientXmlRpcTransportFactory xmlRpcTransportFactory =
                new CommonsHttpClientXmlRpcTransportFactory(serverUrlString, httpClient, debugInfo.isDebugMode());

        ConnectionConfig connectionConfig = getCurrentConnectionConfig(ConnectionConfig.Type.SFA);

        XmlRpcClient xmlRpcClient;
        try {
            xmlRpcClient = new XmlRpcClient(serverUrl, xmlRpcTransportFactory);
            BasicSfaConnection res = new BasicSfaConnection(connectionConfig, xmlRpcTransportFactory, xmlRpcClient);

            if (httpConParts.trustStrategy != null && httpConParts.trustStrategy instanceof SpyingTrustStrategy) {
                ((SpyingTrustStrategy) httpConParts.trustStrategy).setConnection(res);
            }
            return res;
        } catch (Exception e) {
            throw new JFedException("Error creating XmlRpcClient: " + e.getMessage(), e);
        }
    }

    @Override
    public HttpConnection buildHttpConnection() throws JFedException {
        check();

        HttpConParts httpConParts = getConnectionHttpClientAndContext();

        ConnectionConfig connectionConfig = getCurrentConnectionConfig(ConnectionConfig.Type.HTTP);
        BasicHttpConnection res = new BasicHttpConnection(connectionConfig,
                httpConParts.defaultHttpClient, httpConParts.httpContext, httpConParts.preemptiveAuthHttpContext);

        if (httpConParts.trustStrategy != null && httpConParts.trustStrategy instanceof SpyingTrustStrategy) {
            ((SpyingTrustStrategy) httpConParts.trustStrategy).setConnection(res);
        }
        return res;
    }

    @Override
    public ConnectionConfig getCurrentConnectionConfig(ConnectionConfig.Type type) {
        return new ConnectionConfig(serverUrl, debugInfo, proxyInfo, authentication, login, clientCertChain, privateKey, protocol, conTrustStore, handleUntrustedCallback, type);
    }

    @Override
    public void set(ConnectionConfig config) {
        this.serverUrlString = config.getServerUrlString();
        this.debugInfo = config.getDebugInfo();
        this.proxyInfo = config.getProxyInfo();
        this.authentication = config.getAuthentication();
        this.login = config.getLogin();
        this.clientCertChain = config.getClientCertChain();
        this.privateKey = config.getPrivateKey();
        this.protocol = config.getProtocol();
        this.conTrustStore = config.getConTrustStore();
        this.handleUntrustedCallback = config.getHandleUntrustedCallback();
    }


    public static class INSECURE_TRUSTALL_HandleUntrustedCallback implements HandleUntrustedCallback {
        public INSECURE_TRUSTALL_HandleUntrustedCallback() {
            LOG.warn("SECURITY WARNING: constructing INSECURE_TRUSTALL_HandleUntrustedCallback");
        }

        @Override
        public boolean trust(SSLCertificateDownloader.SSLCertificateJFedInfo sSLCertificateJFedInfo) {
            return true;
        }
    }


}
