package be.iminds.ilabt.jfed.lowlevel.stitching;

import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.lowlevel.ServerType;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityFinder;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityListModel;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.util.GeniUrn;
import be.iminds.ilabt.jfed.util.TextUtil;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.w3c.dom.Element;

import javax.annotation.Nullable;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.*;

/**
 * StitchingCallData contains all data about a specific "CreateSLiver" call.
 * This means it contains all data about all hops for the called AM
 * <p/>
 * It is uniquely identified by the AM urn
 * <p/>
 * It also stores:
 * - which VLAN's have failed and which vlan succeeded
 * - how many other failures there where
 * - logs for all calls
 * - current status (simple): sliverExists true or false
 * <p/>
 * It also stores dependencies to other StitchingCallData. To create:
 * - first create it,
 * - then add all StitchingHopData,
 * - then call linkDeps
 */
public class StitchingCallData {
    private final Logger log;

    private final GeniUrn authUrn;
    private final String authUrnString;
    private final SfaAuthority auth;

    private final List<StitchingHopData> hopData = new ArrayList<StitchingHopData>();

    //links (initialised after linkDeps call)
    private List<StitchingCallData> dependencies = new ArrayList<StitchingCallData>();
    private List<StitchingCallData> dependingOnThis = new ArrayList<StitchingCallData>();

    private Set<Integer> unavailableVlans = new HashSet<Integer>();

    private boolean sawAdvertisementRspec = false;

    public static enum State { NONE, ALLOCATED, ALLOCATING, LISTING_RESOURCES, ALLOCATING_BUT_MUST_DELETE, MUST_DELETE, DELETING }

    private State state = State.NONE;

    public StitchingCallData(String authUrnString, String authUrlString, AuthorityListModel authList, AuthorityFinder authFinder, Logger log) {
            if (log == null)
                this.log = LoggerFactory.getLogger(StitchingCallData.class);
            else
                this.log = log;
        this.authUrnString = authUrnString;
        this.authUrn = GeniUrn.parse(authUrnString);
        assert authUrn != null;

        //Do not use this! It HAS to be exact.
//        SfaAuthority aggAuth = authFinder.findByUrn(authUrnString, AuthorityFinder.Purpose.STITCHING);
        SfaAuthority aggAuth = authList.getByUrnExact(authUrnString);

        if (!authUrnString.equals(aggAuth.getUrnString()))
            log.info("Created StitchingCallData for " + authUrnString + "  with new auth urn=" + aggAuth.getUrnString());
        else
            log.info("Created StitchingCallData for " + authUrnString);

        if (aggAuth == null) {
            log.debug("Unknown Hop aggregate \"" + authUrnString + "\". Registering aggregate.");

            if (authUrlString == null)
                throw new RuntimeException("Authority "+authUrnString+" is unknown, and an URL was not provided.");

            if (authUrlString.startsWith("http://")) {
                log.warn("Aggregate URL in hop is http instead of https! " +
                        "This is unsecure, so this connection protocol will never be used. " +
                        "Trying to fix automatically: Changing http to https (might not work). url=" + authUrlString);
                authUrlString = authUrlString.replaceFirst("http", "https");
            }

            Map<ServerType, URL> urls = new HashMap<ServerType, URL>();
            try {
                urls.put(new ServerType(ServerType.GeniServerRole.AM, 2), new URL(authUrlString));
            } catch (MalformedURLException e) {
                throw new RuntimeException("Aggregate URL in hop info is not a valid URL: " + authUrlString, e);
            }
            SfaAuthority newAuth = null;
            try {
                newAuth =
                        new SfaAuthority(authUrnString, authUrnString, null, urls, null/*proxies*/, null, "stitching-hop-aggregate");
            } catch (JFedException e) {
                throw new RuntimeException("Error creating SfaAuthority for hop: cannot continue: " + e.getMessage(), e);
            }
            authList.addAuthority(newAuth); //wil update authList automatically
            aggAuth = newAuth;
            assert aggAuth != null : "Bug: Adding newly created SfaAuthority seems to have failed";
        }
        auth = aggAuth;
    }

    public void addHopData(StitchingHopData hop) {
        assert !hopData.contains(hop);
        hopData.add(hop);
    }

    public StitchingHopData getHopData(String linkName, String hopUrn) {
        for (StitchingHopData hop : hopData) {
            if (hop.getLinkName().equals(linkName) && hop.getHopUrn().equals(hopUrn))
                return hop;
        }
        return null;
    }
    public List<StitchingHopData> getAllHopData() {
        return hopData;
    }

    public void linkDeps(Collection<StitchingCallData> allData) {
        for (StitchingCallData d : allData)
            if (d != this && !dependencies.contains(d))
                for (StitchingHopData othersHop : d.hopData)
                    for (StitchingHopData myHop : hopData) {
                        if (myHop.getDependencies().contains(othersHop)) {
                            if (!dependencies.contains(d)) {
                                dependencies.add(d);
                                d.dependingOnThis.add(this);
                            }
                        }
                    }
    }


    /**
     * Update RSpec info from RSpec received from this authority in reply to a CreateSliver call.
     * <p/>
     * This will change the suggested and available VLAN.
     * <p/>
     * It will only change hops related to this authority!
     */
    public void processManifestRspec(String manifestRspec) {
        Element stitchingElement = StitchingData.extractStitchingElementFromRspec(manifestRspec, log);
        assert stitchingElement != null : "No stitching data in manifest RSpec";
        if (stitchingElement == null)
            return;

        for (StitchingHopData hop : hopData) {
            hop.overwriteWithManifestRspec(manifestRspec, (Element) stitchingElement.cloneNode(true/*deep*/));
        }
    }

    public void exogeniBugWorkAround_registerVlan(int vlan) {
        for (StitchingHopData hop : hopData)
            if (hop.getAggregateUrn().equals(authUrnString))
                hop.exogeniBugWorkAround_registerManifestVlan(vlan);
    }

    /**
     * check if the stitching info in this manifest, is the same as the one that we know is allocated
     *
     * Deprecated: not used anymore. Code may be removed later, if indeed not needed.
     * */
    @Deprecated
    public boolean isSameAs(String rspec) {
        assert isAllocated();
        for (StitchingHopData hop : hopData) {
            boolean same = hop.isSameAs(rspec);
            if (!same)
                return false;
        }
        return true;
    }


    /**
     * @returns true if configuration still possible
     * false if not enough available VLAN's are left
     */
    public boolean setVlanUnavailable(int vlan) {
        //find remaining available VLAN and update hops suggestedVLAN with it
        //keep in account that each hop with same urn but different linkName needs a different VLAN!

        log.trace("setVlanUnavailable({})", vlan);

        unavailableVlans.add(vlan);
        boolean depFailure = false;

        do {
            depFailure = false;

            Map<String, Set<Integer>> vlansUsedByHopUrn = new HashMap<String, Set<Integer>>();
            for (StitchingHopData hop : hopData) {
                log.trace("hop=" + hop.getHopUrn() + "  suggestedVlan=" + hop.getSuggestedVlan());
                Set<Integer> s = vlansUsedByHopUrn.get(hop.getHopUrn());
                if (s == null)
                    s = new HashSet<Integer>();
                Integer suggestedVlan = hop.getSuggestedVlan();
                if (suggestedVlan != null)
                    s.add(suggestedVlan);
                vlansUsedByHopUrn.put(hop.getHopUrn(), s);
            }
            log.trace("vlansUsedByHopUrn=" + vlansUsedByHopUrn);

            List<StitchingHopData> changedHops = new ArrayList<StitchingHopData>();
            for (StitchingHopData hop : hopData) {
                Set<Integer> sameHopVlans = vlansUsedByHopUrn.get(hop.getHopUrn());
                Integer origVlan = hop.getSuggestedVlan();

                Set<Integer> allForbiddenVLANs = new HashSet<Integer>(sameHopVlans);
                if (origVlan != null)
                    allForbiddenVLANs.remove(origVlan); //remove own vlan, will be added again line below if unavailable
                allForbiddenVLANs.addAll(unavailableVlans);

                //find a new VLAN if old is not available anymore
                log.debug("hop " + hop.getHopUrn() + " current suggestion=" + hop.getSuggestedVlan() + " updateSuggestedVlan: forbidden=" + allForbiddenVLANs);
                if (origVlan == null || allForbiddenVLANs.contains(origVlan)) {
                    Integer foundVlan =
                            hop.updateSuggestedVlan(allForbiddenVLANs, null/*no new suggested VLAN, find automatically*/);
                    log.debug("    updated hop VLAN to foundVlan={}", foundVlan);
                    if (foundVlan == null)
                        return false;
                    else
                        sameHopVlans.add(foundVlan);
                    if (origVlan != null && !foundVlan.equals(origVlan))
                        sameHopVlans.remove(origVlan);

                    changedHops.add(hop);
                } else
                    log.debug("   leaving hop VLAN unchanged");
            }


            //hop was updated. now find dependant hops, and update these as well.
            Set<Integer> allForbiddenVLANs = new HashSet<Integer>();
            allForbiddenVLANs.add(vlan);
            while (!changedHops.isEmpty()) {
                List<StitchingHopData> curChangedHops = changedHops;
                changedHops = new ArrayList<StitchingHopData>();
                for (StitchingHopData hop : curChangedHops) {
                    List<StitchingHopData> dependentHops = hop.getDependingOnThis();
                    for (StitchingHopData dependantHop : dependentHops) {
                        Integer foundVlan = dependantHop.updateSuggestedVlan(allForbiddenVLANs, hop.getSuggestedVlan());
                        if (foundVlan == null) {
                            //the dependant hop does not support the VLAN set in its dependency
                            //solution: should blacklist this VLAN, and start again.
                            unavailableVlans.add(hop.getSuggestedVlan());

                            depFailure = true;

                            //check if all vlans on dep are exhausted. In that case, all fails.
                            if (dependantHop.getAvailableVlansString().isEmpty())
                                return false;
                        }
                        //was: if (foundVlan == null) return false; //quick fix: just fail

                        //if this has dependent hops, change those as well.
                        changedHops.add(dependantHop);
                    }
                }
            }

        } while (depFailure);

        //all was successfully updated
        return true;
    }

    public Set<Integer> getUnavailableVlans() {
        return Collections.unmodifiableSet(unavailableVlans);
    }

    public List<Integer> getSuggestedVlans() {
        Set<Integer> res = new HashSet<Integer>();
        for (StitchingHopData hop : hopData)
            res.add(hop.getSuggestedVlan());
        return new ArrayList<Integer>(res);
    }


    public boolean isAllocated() {
        if (getAuth().isFake()) return true;
        return state == State.ALLOCATED;
    }

    public void setState(State newState) {
        this.state = newState;
        if (newState == State.ALLOCATED || newState == State.ALLOCATING)
            for (StitchingCallData dep : dependencies)
                assert dep.isAllocated() || dep.getAuth().isFake();
    }

    public State getState() {
        return state;
    }

    //    public boolean isReady() {
//        if (state != State.NONE) return false;
//
//        for (StitchingCallData dep : dependencies) {
//            if (!dep.isAllocated() && !dep.getAuth().isFake())
//                return false;
//        }
//
//        return true;
//    }
    public boolean areAllDepsReady() {
        for (StitchingCallData dep : dependencies) {
            if (!dep.isAllocated() && !dep.getAuth().isFake())
                return false;
        }

        return true;
    }

    public boolean needsAdvertisementRspec() {
        if (dependencies.isEmpty()) //only needed if we have dependencies.
            return false;
        if (!getAuth().hasSpecialCase(SfaAuthority.SpecialCase.STITCHING_ADVERTISEMENT_RSPEC_UPTODATE))
            return false;
        if (sawAdvertisementRspec)
            return false;
        return true;
    }

    /**
     * parse advertisement stitching info.
     * */
    public void seeAdvertismentRspec(@Nullable String rspec) {
        sawAdvertisementRspec = true;

        if (rspec != null) {
            List<AdvertisementStitchInfo> advertisementStitchInfos = AdvertisementStitchInfo.parse(rspec);
            for (StitchingHopData hop : hopData) {
                AdvertisementStitchInfo bestMatchingInfo = null;
                for (AdvertisementStitchInfo info : advertisementStitchInfos) {

                    if (bestMatchingInfo == null && info.getLinkId().equals(hop.getHopUrn()))
                        bestMatchingInfo = info;
                    //we don't know getHopRemoteUrn , so this is a problem
//                if (info.getLinkId().equals(hop.getHopUrn()) && info.getRemoteLinkId().equals(hop.getHopRemoteUrn()))
//                    bestMatchingInfo = info;
                }
                if (bestMatchingInfo != null)
                    hop.processAdvertismentInfo(bestMatchingInfo.getVlanRangeAvailability());
            }
        }
    }

    /** it is usefull to get the advertisement for everything depending on this, if they support up to date advertisements */
    public boolean areAllDependingOnThisReady() {
        for (StitchingCallData dep : dependingOnThis) {
            if (dep.getAuth().hasSpecialCase(SfaAuthority.SpecialCase.STITCHING_ADVERTISEMENT_RSPEC_UPTODATE) && !dep.sawAdvertisementRspec)
                return false;
        }

        return true;
    }

    public List<StitchingCallData> getDependencies() {
        return Collections.unmodifiableList(dependencies);
    }

    public List<StitchingCallData> getDependingOnThis() {
        return Collections.unmodifiableList(dependingOnThis);
    }

    public GeniUrn getAuthUrn() {
        return authUrn;
    }

    public String getAuthUrnString() {
        return authUrnString;
    }

    public SfaAuthority getAuth() {
        return auth;
    }

    @Override
    public String toString() {
        List<String> depUrns = new ArrayList<String>();
        for (StitchingCallData stitchingCallData : dependencies)
            depUrns.add(stitchingCallData.getAuthUrnString());
        String hopDataString = "[\n";
        for (StitchingHopData stitchingHopData : hopData)
            hopDataString += TextUtil.indent(4, stitchingHopData+",\n");
        hopDataString += "]";

        String res = "\"StitchingCallData\" : {\n";
        res = toStringHelper(res, "authUrnString", authUrnString);
        res = toStringHelper(res, "hopDataString", hopDataString);
        res = toStringHelper(res, "dependencies (urns)", depUrns);
        res += "}\n";
        return res;
    }

    private static String toStringHelper(String res, String objectName, Object objectValue) {
        if (objectValue == null)
            return res;
        return res + TextUtil.indent(4, "\"" + objectName + "\" : \"" + objectValue + "\"")+"\n";
    }




    //same authUrnString is normally same StitchingCallData
    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;

        StitchingCallData that = (StitchingCallData) o;

        assert authUrnString.equals(that.authUrnString) == (this == o);

        if (!authUrnString.equals(that.authUrnString)) return false;

        return true;
    }

    @Override
    public int hashCode() {
        return authUrnString.hashCode();
    }
}
